import logging
from ringo.views.helpers import (
    get_ownership_form,
    get_rendered_ownership_form
)
from ringo.views.request import (
    handle_params,
    handle_history,
    get_return_value,
    get_item_from_request,
    get_item_modul,
    handle_POST_request,
    handle_redirect_on_success
)

log = logging.getLogger(__name__)


def ownership(request, callback=None, renderers=None):
    """Base method to handle requests to change the ownership.

    :request: Current request
    :callback: Current function which is called after the item has been read.
    :returns: Dictionary.
    """
    handle_history(request)
    handle_params(request)
    clazz = request.context.__model__
    item_label = get_item_modul(request, clazz).get_label()
    form = get_ownership_form(request)
    if request.POST:
        item = get_item_from_request(request)
        old_gid = str(item.gid)
        old_uid = str(item.uid)
        new_gid = str(request.params.get('group'))
        new_uid = str(request.params.get('owner'))
        if handle_POST_request(form, request, callback, 'update', renderers):
            if old_uid != new_uid:
                log_msg = u'User {user.login} changed uid of {item_label} '\
                          '{item.id} from {old} -> {new}'\
                          .format(item_label=item_label, item=item, 
                                  user=request.user, old=old_uid, new=new_uid)
                log.info(log_msg)
            if old_gid != new_gid:
                log_msg = u'User {user.login} changed gid of {item_label} '\
                          '{item.id} from {old} -> {new}'\
                          .format(item_label=item_label, item=item, 
                                  user=request.user, old=old_gid, new=new_gid)
                log.info(log_msg)
            return handle_redirect_on_success(request)
    rvalues = get_return_value(request)
    rvalues['form'] = get_rendered_ownership_form(request)
    return rvalues



import logging
from string import Template
import json
from pyramid.events import NewRequest
from pyramid.i18n import get_localizer, TranslationStringFactory
from ringo.lib.helpers import literal


log = logging.getLogger(__name__)

translators = []
translators.append(TranslationStringFactory('formbar'))
translators.append(TranslationStringFactory('ringo'))
"""The translators variable holds a list of available
TranslationStringFactory. Usually there is one factory per i18n domain
which needs to be translated. On translation the availabe translation
strings will be iterated until one of them returns a translated string.
On default there is only two translators available for ringo. The base
ringo translator and the formbar translator. But derived applications
can append their own translation factory."""

def setup_translation(config):
    config.add_translation_dirs('ringo:locale/')
    config.add_translation_dirs('formbar:locale/')
    config.add_subscriber(set_request_locale, NewRequest)
    config.add_subscriber(add_localizer, NewRequest)
    return config

def set_request_locale(event):
    """Will set the loacle of the request depending on the users
    accept_language setting in the browser. The locale will be used for
    localisation."""
    if not event.request.accept_language:
        return
    accepted = event.request.accept_language
    event.request._LOCALE_ = accepted.best_match(('en', 'fr', 'de'), 'en')

def add_localizer(event):
    request = event.request
    localizer = request.localizer

    def auto_translate(string, default=None, mapping=None):
        """The translation function will iterate over the available
        TranslationStringFactorys in the translators varibale and try to
        translate the given string. TranslationsStrings will be iterated
        in reversed order and iteration stops as soon as the
        TranslationsString returns a string different from the given
        string (msgid)"""
        # An empty string leads to returning the header of the PO file.
        # So if no string to translate is provided return an empty string.
        if not string:
            return string
        if mapping is None:
            mapping = {}
        for _ in translators[::-1]:
            ts = localizer.translate(_(string,
                                       default=default))
            if ts != string:
                break
        translated = Template(ts).safe_substitute(mapping)
        # If the origin string is a "literal" object (has __html__
        # attribute, then return a literal again.
        if hasattr(string, '__html__'):
            return literal(translated)
        return translated

    request.localizer = localizer
    request.translate = auto_translate


def locale_negotiator(request):
    accepted = request.accept_language
    return accepted.best_match(('en', 'fr', 'de'), 'en')


def extract_i18n_tableconfig(fileobj, keywords, comment_tags, options):
    """Extract messages from JSON table configuration files.
    :param fileobj: the file-like object the messages should be extracted
                    from
    :param keywords: a list of keywords (i.e. function names) that should
                     be recognized as translation functions
    :param comment_tags: a list of translator tags to search for and
                         include in the results
    :param options: a dictionary of additional options (optional)
    :return: an iterator over ``(lineno, funcname, message, comments)``
             tuples
    :rtype: ``iterator``
    """
    config = json.load(fileobj)
    # FIXME: Fix linenumbering. No real linenummer, just iterate somehow
    lineno = 0
    for key, tc in config.iteritems():
        lineno += 1
        for col in tc.get('columns'):
            lineno += 1
            # "_" is one of the default keywords which marks a string
            # for extraction. As the json file does not have any
            # keywords. Set a dummy funcname here.
            yield (lineno,
                   "_",
                   col.get('label'),
                   ["Label for %s column in %s table config"
                    % (col.get('name'), key)])
        for col in tc.get('filters', []):
            lineno += 1
            # "_" is one of the default keywords which marks a string
            # for extraction. As the json file does not have any
            # keywords. Set a dummy funcname here.
            yield (lineno,
                   "_",
                   col.get('label'),
                   ["Label for %s filter in %s table config"
                    % (col.get('field'), key)])
        for col in tc.get('filters', []):
            lineno += 1
            # "_" is one of the default keywords which marks a string
            # for extraction. As the json file does not have any
            # keywords. Set a dummy funcname here.
            yield (lineno,
                   "_",
                   col.get('expr'),
                   ["Expr for %s filter in %s table config"
                    % (col.get('field'), key)])

#@subscriber(NewRequest)
#def setAcceptedLanguagesLocale(event):
#    request = event.request
#    if not request.accept_language:
#        return
#    event.request._LOCALE_ = locale_negotiator(event.request)



import sys
import logging
import query
from zope.sqlalchemy import ZopeTransactionExtension
from pyramid.events import NewRequest
from sqlalchemy import engine_from_config
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy import exc
from sqlalchemy import event
from sqlalchemy.pool import Pool, StaticPool

from ringo.lib.sql.cache import regions, init_cache

log = logging.getLogger(__name__)

# Session initialisation
########################
# scoped_session.  Apply our custom CachingQuery class to it,
# using a callable that will associate the dictionary
# of regions with the Query.
DBSession = scoped_session(
                sessionmaker(
                    query_cls=query.query_callable(regions),
                )
            )

NTDBSession = scoped_session(
                sessionmaker(
                    query_cls=query.query_callable(regions)
                )
            )
testsession = None


# Pessimisitic detection of disconnects of connections in the connection pool
# (E.g the database server restarts because of maintenance. Per checkout
# of a connection from the pool we will do a extra SQL-Query. In case
# this fails because of the disconnection the connection pool will be
# invalidated and the connections will be reestablished. See
# http://docs.sqlalchemy.org/en/latest/core/pooling.html#disconnect-handling-pessimistic
# for for details.
@event.listens_for(Pool, "checkout")
def ping_connection(dbapi_connection, connection_record, connection_proxy):
    cursor = dbapi_connection.cursor()
    try:
        cursor.execute("SELECT 1")
    except:
        log.info("Possible DB disconnect detected! "
                 "Reestablishing the connection to the DB.")
        # optional - dispose the whole pool
        # instead of invalidating one at a time
        connection_proxy._pool.dispose()

        # raise DisconnectionError - pool will try
        # connecting again up to three times before raising.
        raise exc.DisconnectionError()
    cursor.close()


def setup_db_engine(settings):
    cachedir = settings.get("db.cachedir")
    regions = []
    for region in settings.get("db.cacheregions", "").split(" "):
        regions.append(region.split(":"))
    if cachedir:
        init_cache(cachedir, regions)
    if settings.get("app.mode") == "testing":
        return engine_from_config(settings, 'sqlalchemy.',
                                  poolclass=StaticPool)
    else:
        return engine_from_config(settings, 'sqlalchemy.')


def setup_db_session(engine, settings=None):
    # Onyl use ZopeTransactionExtension if not in testmode to prevent
    # autocommits after each request.
    if not settings:
        settings = {}
    if settings.get("app.mode") == "testing":
        DBSession.configure(bind=engine)
    else:
        DBSession.configure(bind=engine, extension=ZopeTransactionExtension())
    NTDBSession.configure(bind=engine)


# Session initialisation
########################
def setup_session_on_request(config):
    if config.registry.settings.get("app.mode") == "testing":
        config.add_subscriber(add_test_session_to_request, NewRequest)
        # Add exception view to rollback the database in case of
        # errors in testmode. Adding a view here feels wrong, but it is
        # the currently  only known way to to handle exceptions.
        config.add_view(rollback_on_exception, context=Exception)
    else:
        config.add_subscriber(add_session_to_request, NewRequest)


def rollback_on_exception(exc, request):
    """If the application is in testing mode, this view will be called
    on every exeption and takes care that the current transaction is
    rolled back."""
    request.db.rollback()
    request.db.close()
    # Reraise the origin exception without loosing the stacktrace.
    raise sys.exc_info()[0], sys.exc_info()[1], sys.exc_info()[2]


def add_test_session_to_request(event):
    global testsession
    request = event.request
    if testsession is not None:
        request.db = testsession
        request._active_testcase = True
    elif request.params.get("_testcase") == "begin":
        testsession = DBSession()
        request.db = testsession
        request._active_testcase = True
    else:
        request.db = DBSession()
        request._active_testcase = False
    request.add_finished_callback(close_test_session)


def close_test_session(request):
    global testsession
    if testsession is not None:
        if request.params.get("_testcase") == "end":
            log.debug("End Testcase")
            request.db.rollback()
            request.db.close()
            testsession = None
        else:
            request.db.flush()
    else:
        request.db.commit()


def add_session_to_request(event):
    request = event.request
    request.db = DBSession()
    request.add_finished_callback(close_session)


def close_session(request):
    if request.db:
        request.db.close()



import os
from invoke import run
import pkg_resources

def get_package_location(name):
    return pkg_resources.get_distribution(name).location

def get_package_name(config_file):
    result = run("sed -n 's|^use = egg:\\([[:graph:]]*\\).*|\\1|p' %s | head -1" % config_file, hide="out")
    return result.stdout.strip()


def get_db(config_file):
    result = run("sed -n 's|^sqlalchemy.url.*//.*@.*/\\([[:graph:]]*\\).*|\\1|p' %s" % config_file, hide="out")
    return result.stdout.strip()


def get_fixtures(appname, path=None):
    if path:
        try:
            result = run("ls %s/*.json" % path, hide="out").stdout.strip()
        except Exception as e:
            return []
    else:
        apppath = os.path.join(get_package_location(appname), appname)
        result = run("ls %s/fixtures/*.json"
                     % apppath, hide="out").stdout.strip()
    fixtures = []
    for fixture in sorted(result.split("\n")):
        fixtures.append((fixture, "_".join(fixture.split("_")[1:]).split(".")[0]))
    return fixtures



import logging
from ringo.lib.imexport import (
    JSONExporter,
    CSVExporter,
    XLSXExporter
)
from ringo.lib.renderer import (
    ExportDialogRenderer
)
from ringo.views.request import (
    handle_params,
    handle_history,
    is_confirmed
)
from ringo.views.helpers import get_item_from_request
from ringo.views.base.list_ import set_bundle_action_handler

log = logging.getLogger(__name__)


def export(request):
    handle_history(request)
    handle_params(request)
    item = get_item_from_request(request)
    return _handle_export_request(request, [item])


def _handle_export_request(request, items, callback=None):
    """Helper function to handle the export request. This function
    provides the required logic to show the export configuration dialog
    and returning the exported items. It is called when exporting a
    single item or when exporting multiple items in a bundle."""
    clazz = request.context.__model__
    renderer = ExportDialogRenderer(request, clazz)
    form = renderer.form
    if (request.method == 'POST'
       and is_confirmed(request)
       and form.validate(request.params)):
        # Setup exporter
        ef = form.data.get('format')
        if ef == "json":
            exporter = JSONExporter(clazz)
        elif ef == "csv":
            exporter = CSVExporter(clazz)
        elif ef == "xlsx":
            exporter = XLSXExporter(clazz)
        export = exporter.perform(items)
        # Build response
        resp = request.response
        resp.content_type = str('application/%s' % ef)
        resp.content_disposition = 'attachment; filename=export.%s' % ef
        resp.body = export
        return resp
    else:
        # FIXME: Get the ActionItem here and provide this in the Dialog to get
        # the translation working (torsten) <2013-07-10 09:32>
        rvalue = {}
        rvalue['dialog'] = renderer.render(items)
        return rvalue

set_bundle_action_handler("export", _handle_export_request)



"""Initial Ringo 1.0 migration

Revision ID: 57c6ab33ad2f
Revises: None
Create Date: 2015-02-10 10:49:34.997715

"""

# revision identifiers, used by Alembic.
revision = '57c6ab33ad2f'
down_revision = None

from alembic import op
import sqlalchemy as sa


UPGRADE = """
"""
DOWNGRADE = """"""


def iter_statements(stmts):
    for st in [x for x in stmts.split('\n') if x]:
        op.execute(st)


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table('roles',
    sa.Column('uuid', sa.CHAR(length=32), nullable=False),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('label', sa.String(), nullable=False),
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('description', sa.Text(), nullable=False),
    sa.Column('admin', sa.Boolean(), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('label'),
    sa.UniqueConstraint('name'),
    sa.UniqueConstraint('uuid')
    )
    op.create_table('user_settings',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('settings', sa.Text(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('versions',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('values', sa.Text(), nullable=False),
    sa.Column('author', sa.String(), nullable=False),
    sa.Column('date', sa.DateTime(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('usergroups',
    sa.Column('uuid', sa.CHAR(length=32), nullable=False),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('description', sa.String(), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('name'),
    sa.UniqueConstraint('uuid')
    )
    op.create_table('users',
    sa.Column('uuid', sa.CHAR(length=32), nullable=False),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('login', sa.String(), nullable=False),
    sa.Column('password', sa.String(), nullable=False),
    sa.Column('activated', sa.Boolean(), nullable=True),
    sa.Column('activation_token', sa.String(), nullable=False),
    sa.Column('gid', sa.Integer(), nullable=True),
    sa.Column('sid', sa.Integer(), nullable=True),
    sa.Column('last_login', sa.DateTime(), nullable=True),
    sa.ForeignKeyConstraint(['gid'], ['usergroups.id'], ),
    sa.ForeignKeyConstraint(['sid'], ['user_settings.id'], ),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('login'),
    sa.UniqueConstraint('uuid')
    )
    op.create_table('nm_usergroup_roles',
    sa.Column('gid', sa.Integer(), nullable=True),
    sa.Column('rid', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['gid'], ['usergroups.id'], ),
    sa.ForeignKeyConstraint(['rid'], ['roles.id'], )
    )
    op.create_table('modules',
    sa.Column('uuid', sa.CHAR(length=32), nullable=False),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('clazzpath', sa.String(), nullable=False),
    sa.Column('label', sa.String(), nullable=False),
    sa.Column('label_plural', sa.String(), nullable=False),
    sa.Column('description', sa.Text(), nullable=False),
    sa.Column('str_repr', sa.String(), nullable=False),
    sa.Column('display', sa.String(), nullable=False),
    sa.Column('gid', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['gid'], ['usergroups.id'], ),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('clazzpath'),
    sa.UniqueConstraint('name'),
    sa.UniqueConstraint('uuid')
    )
    op.create_table('actions',
    sa.Column('uuid', sa.CHAR(length=32), nullable=False),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('mid', sa.Integer(), nullable=True),
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('url', sa.String(), nullable=False),
    sa.Column('icon', sa.String(), nullable=False),
    sa.Column('description', sa.Text(), nullable=False),
    sa.Column('bundle', sa.Boolean(), nullable=False),
    sa.Column('display', sa.String(), nullable=False),
    sa.Column('permission', sa.String(), nullable=False),
    sa.ForeignKeyConstraint(['mid'], ['modules.id'], ),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('uuid')
    )
    op.create_table('nm_user_roles',
    sa.Column('uid', sa.Integer(), nullable=True),
    sa.Column('rid', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['rid'], ['roles.id'], ),
    sa.ForeignKeyConstraint(['uid'], ['users.id'], )
    )
    op.create_table('password_reset_requests',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('uid', sa.Integer(), nullable=True),
    sa.Column('created', sa.DateTime(), nullable=True),
    sa.Column('token', sa.String(), nullable=True),
    sa.ForeignKeyConstraint(['uid'], ['users.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('profiles',
    sa.Column('uuid', sa.CHAR(length=32), nullable=False),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('first_name', sa.String(), nullable=True),
    sa.Column('last_name', sa.String(), nullable=True),
    sa.Column('gender', sa.Integer(), nullable=True),
    sa.Column('birthday', sa.Date(), nullable=True),
    sa.Column('address', sa.Text(), nullable=True),
    sa.Column('phone', sa.String(), nullable=True),
    sa.Column('email', sa.String(), nullable=True),
    sa.Column('web', sa.String(), nullable=True),
    sa.Column('uid', sa.Integer(), nullable=True),
    sa.Column('gid', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['gid'], ['usergroups.id'], ),
    sa.ForeignKeyConstraint(['uid'], ['users.id'], ),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('uuid')
    )
    op.create_table('nm_user_usergroups',
    sa.Column('uid', sa.Integer(), nullable=True),
    sa.Column('gid', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['gid'], ['usergroups.id'], ),
    sa.ForeignKeyConstraint(['uid'], ['users.id'], )
    )
    op.create_table('forms',
    sa.Column('created', sa.DateTime(), nullable=True),
    sa.Column('updated', sa.DateTime(), nullable=True),
    sa.Column('uuid', sa.CHAR(length=32), nullable=False),
    sa.Column('review_state_id', sa.Integer(), nullable=True),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('category', sa.Integer(), nullable=True),
    sa.Column('title', sa.String(), nullable=True),
    sa.Column('description', sa.Text(), nullable=True),
    sa.Column('definition', sa.Text(), nullable=True),
    sa.Column('mid', sa.Integer(), nullable=True),
    sa.Column('uid', sa.Integer(), nullable=True),
    sa.Column('gid', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['gid'], ['usergroups.id'], ),
    sa.ForeignKeyConstraint(['mid'], ['modules.id'], ),
    sa.ForeignKeyConstraint(['uid'], ['users.id'], ),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('uuid')
    )
    op.create_table('nm_action_roles',
    sa.Column('aid', sa.Integer(), nullable=True),
    sa.Column('rid', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['aid'], ['actions.id'], ),
    sa.ForeignKeyConstraint(['rid'], ['roles.id'], )
    )
    ### end Alembic commands ###
    iter_statements(UPGRADE)


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('nm_action_roles')
    op.drop_table('forms')
    op.drop_table('nm_user_usergroups')
    op.drop_table('profiles')
    op.drop_table('password_reset_requests')
    op.drop_table('nm_user_roles')
    op.drop_table('actions')
    op.drop_table('modules')
    op.drop_table('nm_usergroup_roles')
    op.drop_table('users')
    op.drop_table('usergroups')
    op.drop_table('versions')
    op.drop_table('user_settings')
    op.drop_table('roles')
    ### end Alembic commands ###
    iter_statements(DOWNGRADE)



from pyramid.view import view_config
from pyramid.httpexceptions import HTTPFound


@view_config(route_name='start_test_case')
def begin(request):
    if request.params.get("_testcase"):
        history = request.session['history']
        url = history.pop()
    else:
        url = request.route_path("start_test_case",
                                 _query={"_testcase": "begin"})
    return HTTPFound(url)


@view_config(route_name='stop_test_case')
def end(request):
    if request.params.get("_testcase"):
        history = request.session['history']
        url = history.pop()
    else:
        url = request.route_path("stop_test_case",
                                 _query={"_testcase": "end"})
    return HTTPFound(url)



import os
import logging
import pkg_resources
import transaction
from pyramid.events import NewRequest
from ringo.lib import helpers
from ringo.lib.extension import check_unregister_modul
from ringo.lib.sql.db import DBSession, NTDBSession
from ringo.lib.helpers import get_action_routename
from ringo.lib.form import get_formbar_css, get_formbar_js
from ringo.model.modul import ModulItem
from ringo.model import extensions
from ringo.model.mixins import Mixin
from ringo.model.base import get_item_list
from ringo.resources import get_resource_factory
from ringo.views.base import (
    web_action_view_mapping,
    rest_action_view_mapping,
    get_action_view
)

log = logging.getLogger(__name__)

def preload_modules(event):
    """Preload all modules on each request and put them into the request
    cache for the modules to save additional SQL queries as the modules
    are needed anyway."""
    for modul in get_item_list(event.request, ModulItem):
        event.request.cache_item_modul.set(modul.id, modul)

def setup(config):
    """Setup method which is called on application initialition and
    takes care that many ringo specific things are setup correctly."""

    # Trigger loading the static files of formbar and make them
    # available in global vars.
    # TODO: Move this call into a better place. (ti) <2015-07-28 13:19> 
    get_formbar_css() # -> formbar_css_filenames
    get_formbar_js() # -> formbar_js_filenames

    setup_extensions(config)
    setup_modules(config)
    config.include('ringo.lib.i18n.setup_translation')
    config.include('ringo.lib.sql.db.setup_session_on_request')
    config.include('ringo.lib.renderer.setup_render_globals')
    config.include('ringo.lib.security.setup_ringo_security')
    config.include('ringo.lib.cache.setup_cache')
    config.add_subscriber(preload_modules, NewRequest)


def setup_extensions(config):
    """Will setup the configured extensions. This included initialising
    new extension and removing extension which aren't configured any
    longer in the application."""
    # First initialise extenstion
    for extension in extensions:
        config.include(extension)
    # Now delete all extensions which are not configured.
    for modul in DBSession.query(ModulItem).all():
        check_unregister_modul(modul, extensions)


def setup_modules(config):
    """Will iterate over all configured modules in the application.
    Configured modules are loaded from database. For each module it will
    call the setup_modul method."""
    for modul in NTDBSession.query(ModulItem).all():
        setup_modul(config, modul)


def _setup_web_action(config, action, clazz, view_mapping):
    """Setup a route and a view for given action and clazz.
    The routes will have the follwoing following name and url:

    * Name: $modulname-$actionname
    * Url:  $modulname/$actionurl

    Note, that the actionname can be configured only as admin.

    Further a clazz specific factory will be added to the route which is
    later used to setup the ACL of items of the modul.

    :config: Configuration
    :action: Action item
    :clazz: clazz item
    :view_mapping: Dictionary with action items
    :returns: @todo
    """
    name = clazz.__tablename__
    action_name = action.name.lower()
    route_name = get_action_routename(clazz, action_name)
    route_url = "%s/%s" % (name, action.url)
    log.debug("Adding WEB route: %s, %s" % (route_name, route_url))
    config.add_route(route_name, route_url,
                     factory=get_resource_factory(clazz))
    settings = config.registry.settings
    http_cache = settings.get('security.page_http_cache', '0')
    view_func = get_action_view(view_mapping,
                                  action_name,
                                  name)
    if view_func:
        if action_name == "delete":
            template = "confirm"
            renderer = '/default/%s.mako' % template
        elif action_name == "download":
            renderer = None
        else:
            template = action_name
            renderer = '/default/%s.mako' % template
        config.add_view(view_func,
                        route_name=route_name,
                        renderer=renderer,
                        permission=action.permission or action_name,
                        http_cache=int(http_cache))
    ## Add bundle action.
    if action_name == "list":
       action_name = "bundle"
       route_name = "%s-%s" % (name, action_name)
       route_url = "%s/%s" % (name, action_name)
       view_func = get_action_view(view_mapping,
                                     action_name,
                                     name)
       log.debug("Adding route: %s, %s" % (route_name, route_url))
       config.add_route(route_name, route_url,
                        factory=get_resource_factory(clazz))
       config.add_view(view_func, route_name=route_name,
                       renderer='/default/bundle.mako',
                       permission='list')
    ## Add permission action.
    if action_name == "read":
        action_name = "ownership"
        route_name = "%s-%s" % (name, action_name)
        route_url = "%s/%s/{id}" % (name, action_name)
        view_func = get_action_view(view_mapping,
                                    action_name,
                                    name)
        log.debug("Adding route: %s, %s" % (route_name, route_url))
        config.add_route(route_name, route_url,
                         factory=get_resource_factory(clazz))
        config.add_view(view_func, route_name=route_name,
                        renderer='/default/update.mako',
                        permission='read')



def _setup_rest_action(config, action, clazz, view_mapping):
    """Setup a route and a view for given action and clazz.
    The routes will have the follwoing following name and url:

    * Name: $modulname-$actionname
    * Url:  $modulname/$actionurl

    Note, that the actionname can be configured only as admin.

    Further a clazz specific factory will be added to the route which is
    later used to setup the ACL of items of the modul.

    :config: Configuration
    :action: Action item
    :clazz: clazz item
    :view_mapping: Dictionary with action items
    :returns: @todo
    """
    action_method_mapping = {
        "list": "GET",
        "create": "POST",
        "read": "GET",
        "update": "PUT",
        "delete": "DELETE"
    }
    name = clazz.__tablename__
    action_name = action.name.lower()
    view_func = get_action_view(view_mapping,
                                  action_name,
                                  name)
    if not view_func:
        return
    route_name = get_action_routename(clazz, action_name,
                                      prefix="rest")
    tmpurl = action.url.split("/")
    if len(tmpurl) > 1:
        route_url = "rest/%s/%s" % (name, tmpurl[1])
    else:
        route_url = "rest/%s" % (name)
    log.debug("Adding REST route: %s, %s" % (route_name, route_url))
    method = action_method_mapping[action_name]
    config.add_route(route_name, route_url,
                     request_method=method,
                     factory=get_resource_factory(clazz))
    log.debug("Adding REST view: %s, %s, %s" % (view_func, route_name, method))
    settings = config.registry.settings
    http_cache = settings.get('security.page_http_cache', '0')
    config.add_view(view_func,
                    route_name=route_name,
                    request_method=method,
                    renderer='json',
                    permission=action.permission or action_name,
                    http_cache=int(http_cache))


def setup_modul(config, modul):
    """Setup routes and views for the activated actions of the given
    model of the modul.

    :config: Pylons config instance
    :modul: The module for which the new routes will be set up.
    """
    clazz = helpers.dynamic_import(modul.clazzpath)
    log.info("Setup modul '%s'" % modul.name)

    # Reload modul
    old_actions = list(a.url for a in modul.actions)

    for bclazz in clazz.__bases__:
        if issubclass(bclazz, Mixin):
            for action in bclazz.get_mixin_actions():
                if not action.url in old_actions:
                    action.mid = clazz._modul_id
                    modul.actions.append(action)
                    NTDBSession.add(action)
    NTDBSession.commit()

    for action in modul.actions:
        _setup_web_action(config, action, clazz, web_action_view_mapping)
        _setup_rest_action(config, action, clazz, rest_action_view_mapping)



import os
import logging
from pyramid.config import Configurator
from pyramid_beaker import session_factory_from_settings

from ringo.resources import get_resource_factory
from ringo.lib.i18n import locale_negotiator
from ringo.lib.sql.db import setup_db_session, setup_db_engine
from ringo.model import Base
from ringo.model.user import User, Usergroup

log = logging.getLogger(__name__)


def main(global_config, **settings):
    """ This function returns a Pyramid WSGI application.
    """
    # Setup two db sessions. One using transactions (default) and one
    # without transactions.
    databaseurl = os.environ.get('DATABASE_URL')
    if databaseurl:
        settings['sqlalchemy.url'] = databaseurl
        log.info("Using database url from ENV: %s" % databaseurl)
    engine = setup_db_engine(settings)
    setup_db_session(engine, settings)
    Base.metadata.bind = engine
    config = Configurator(settings=settings,
                          locale_negotiator=locale_negotiator)
    config.set_session_factory(session_factory_from_settings(settings))
    config.include('ringo')
    return config.make_wsgi_app()


def includeme(config):
    log.info('Setup of Ringo...')
    # Configure pyramid modules
    config.include('pyramid_beaker')
    config.include('pyramid_mako')
    config.include('ringo.config.setup')
    config = setup_static_views(config)
    config = setup_routes(config)
    config.scan()
    log.info('OK :) Setup of Ringo finished.')


def setup_static_views(config):
    settings = config.registry.settings
    http_cache = settings.get('security.static_http_cache', '3600')
    config.add_static_view('formbar-static',
                           path='formbar:static',
                           cache_max_age=int(http_cache))
    config.add_static_view('ringo-static',
                           path='ringo:static',
                           cache_max_age=int(http_cache))
    return config


def setup_routes(config):
    """Function which will setup the routes of the ringo application"""

    # SINGLE PAGES
    ##############
    config.add_route('start_test_case', '_test_case/start')
    config.add_route('stop_test_case', '_test_case/stop')
    config.add_route('login', 'auth/login')
    config.add_route('register_user', 'auth/register_user')
    config.add_route('confirm_user', 'auth/confirm_user/{token}')
    config.add_route('forgot_password', 'auth/forgot_password')
    config.add_route('reset_password', 'auth/reset_password/{token}')
    config.add_route('logout', 'auth/logout')
    config.add_route('autologout', 'auth/autologout')
    config.add_route('accountdisabled', 'auth/disabled')
    config.add_route('version', 'version')
    config.add_route('contact', 'contact')
    config.add_route('about', 'about')
    config.add_route('home', '/')

    # Helpers
    #########
    config.add_route('set_current_form_page', 'set_current_form_page')
    config.add_route('users-changepassword',
                     'users/changepassword/{id}',
                     factory=get_resource_factory(User))
    config.add_route('users-removeaccount',
                     'users/removeaccount/{id}',
                     factory=get_resource_factory(User))
    config.add_route('users-accountremoved',
                     'users/accountremoved')
    config.add_route('usergroups-setstandin',
                     'usergroups/setstandin/{id}',
                     factory=get_resource_factory(Usergroup))
    config.add_route('rules-evaluate', 'rest/rule/evaluate')
    config.add_route('form-render', 'rest/form/render')
    config.add_route('keepalive', 'rest/keepalive')
    return config



def clickjacking_factory(handler, registry):
    def clickjacking_tween(request):
        settings, response = registry.settings, request.response
        HEADER = 'X-Frame-Options'
        if not response.headers.get(HEADER, None):
            option = settings.get('x_frame_options', 'DENY').upper()
            response.headers[HEADER] = option
        return handler(request)
    return clickjacking_tween



"""Removed nm_usergroup_roles table

Revision ID: 4ff2bc4eb9bf
Revises: 57c6ab33ad2f
Create Date: 2015-05-10 20:28:01.388673

"""

# revision identifiers, used by Alembic.
revision = '4ff2bc4eb9bf'
down_revision = '57c6ab33ad2f'

from alembic import op
import sqlalchemy as sa


UPGRADE = """"""
DOWNGRADE = """"""


def iter_statements(stmts):
    for st in [x for x in stmts.split('\n') if x]:
        op.execute(st)


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('nm_usergroup_roles')
    ### end Alembic commands ###
    iter_statements(UPGRADE)


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table('nm_usergroup_roles',
    sa.Column('gid', sa.INTEGER(), nullable=True),
    sa.Column('rid', sa.INTEGER(), nullable=True),
    sa.ForeignKeyConstraint(['gid'], [u'usergroups.id'], ),
    sa.ForeignKeyConstraint(['rid'], [u'roles.id'], )
    )
    ### end Alembic commands ###
    iter_statements(DOWNGRADE)



import logging
from pyramid.httpexceptions import HTTPBadRequest
from formbar.config import (
    Config,
    parse
)
from formbar.form import Form
from ringo.lib.helpers import (
        get_item_modul,
        get_app_url
)
from ringo.lib.form import (
    get_eval_url,
    get_form_config,
    get_ownership_form as _get_ownership_form
)
from ringo.lib.i18n import locale_negotiator
from ringo.lib.security import has_role
from ringo.lib.renderer import add_renderers
from ringo.model.user import Usergroup
from ringo.model.form import Form as BlobformForm
from ringo.model.mixins import (
    Owned,
    Versioned,
    Blobform
)

log = logging.getLogger(__name__)


def get_blobform_config(request, item, formname):
    """Helper function used in the create_ method to setup the create
    forms for blogform items. To create a new blogform item the
    creation is done in three steps:

    1. Stage 1: The user selects a form from a list
    2. Stage 2: The create dialog is rendered with the selected form
    3. Stage 3: The item is validated and saved.

    :request: current request
    :item: item to build the form
    :formname: name of the form in the formconfig
    :returns: formconfig, item used to build a form.

    """
    # First check if the fid parameter is provided
    fid = request.params.get('fid') or item.fid
    blobform = request.params.get('blobforms')
    if fid:
        log.debug("Stage 3: User has submitted data to create a new item")
        setattr(item, 'fid', fid)
        formfactory = BlobformForm.get_item_factory()
        formconfig = Config(parse(formfactory.load(fid).definition))
        return item, formconfig.get_form(formname)
    elif blobform:
        log.debug("Stage 2: User has selected a blobform %s " % blobform)
        setattr(item, 'fid', blobform)
        formfactory = BlobformForm.get_item_factory()
        formconfig = Config(parse(formfactory.load(blobform).definition))
        return item, formconfig.get_form(formname)
    else:
        log.debug("Stage 1: User is selecting a blobform")
        modul = get_item_modul(request, item)
        formconfig = get_form_config(modul, "blobform")
        return modul, formconfig


def get_rendered_ownership_form(request):
    """Returns the rendered ownership form for the item in the current
    request. If the item is not an instance of Owned, than an empty
    string is returned.

    Changing the owner of the item will only be available for users with
    a administrative role and update permissions on the current item.
    Changing the group is restricted to the groups the user is member if
    the user has not an administrative role.
    """

    def _has_administrational_role(modul, user):
        for action in modul.actions:
            if action.name == "Update":
                for role in action.roles:
                    if role.admin and has_role(user, role.name):
                        return True
        return False

    item = get_item_from_request(request)
    form = get_ownership_form(request)
    modul = get_item_modul(request, item)
    usergroup_modul = get_item_modul(request, Usergroup)
    _groups = [unicode(g.name) for g in request.user.groups]
    _admin = (_has_administrational_role(modul, request.user)
              or has_role(request.user, "admin")
              or _has_administrational_role(usergroup_modul, request.user))
    values = {"_admin": _admin,
              "_groups": _groups}
    if isinstance(item, Owned):
        return form.render(values=values)
    else:
        return ""


def get_ownership_form(request):
    item = get_item_from_request(request)
    db = request.db
    csrf_token = request.session.get_csrf_token()
    url_prefix = get_app_url(request)

    # Check if the form is rendered as readonly form.
    if has_role(request.user, "admin"):
        readonly = False
    elif isinstance(item, Owned) and item.is_owner(request.user):
        readonly = False
    else:
        readonly = True

    return _get_ownership_form(item, db, csrf_token, get_eval_url(),
                               readonly, url_prefix,
                               locale=locale_negotiator(request),
                               translate=request.translate)


def render_item_form(request, form, values=None, validate=True):
    """Returns the rendered item form for the item in the current
    request."""
    if not values:
        values = {}
    item = form._item
    if isinstance(item, Versioned):
        previous_values = item.get_previous_values(author=request.user.login)
    else:
        previous_values = {}
    # Validate the form to generate the warnings if the form has not
    # been alreaded validated.
    if validate and not form.validated:
        form.validate(None)
    clazz = request.context.__model__
    page = get_current_form_page(clazz, request)
    # Add ringo specific values into the renderered form
    return form.render(page=page, values=values,
                       previous_values=previous_values)


def get_item_form(name, request, renderers=None, validators=None, values=None):
    """Will return a form for the given item

    :name: Name of the form
    :request: Current request
    :renderers: Dictionary of external renderers which should be used
                for renderering some form elements.
    :validators: List of external formbar validators which should be
                 added to the form for validation
    :values: Dictionary with external values to prefill the form or add
    addional values for rule evaluation.
    :returns: Form
    """
    if renderers is None:
        renderers = {}
    if validators is None:
        validators = []
    if values is None:
        values = {}
    item = get_item_from_request(request)
    renderers = add_renderers(renderers)
    clazz = request.context.__model__
    name = request.session.get("%s.form" % clazz) or name

    ## handle blobforms
    if isinstance(item, Blobform):
        # TODO: Why not use the get_form_config method here. This can
        # handle Blobforms and usual form configs. () <2014-08-26 22:21>

        item, formconfig = get_blobform_config(request, item, name)
    else:
        formconfig = get_form_config(item, name)

    form = Form(formconfig, item, request.db,
                translate=request.translate,
                renderers=renderers,
                change_page_callback={'url': 'set_current_form_page',
                                      'item': clazz.__tablename__,
                                      'itemid': item.id},
                request=request,
                csrf_token=request.session.get_csrf_token(),
                eval_url=get_eval_url(),
                url_prefix=get_app_url(request),
                locale=locale_negotiator(request),
                values=values)
    # Add validators
    for validator in validators:
        form.add_validator(validator)
    return form


def get_current_form_page(clazz, request):
    """Returns the id of the currently selected page. The currently
    selected page is saved in the session. If there is no saved value
    then the the first page is returned

    :clazz: The clazz for which the form is displayed
    :request: Current request
    :returns: id of the currently selected page. Default: 1
    """
    itemid = request.matchdict.get('id')
    item = clazz.__tablename__
    page = request.session.get('%s.%s.form.page' % (item, itemid))
    if page:
        return int(page)
    else:
        return 1


def get_next_form_page(form, current_page):
    """Returns the id of the next formpage, Returns current_page if
    there is no next page.

    :form: A formbar form instance
    :current_page: id of the current page:
    :returns: ID of the next page in the form
    """
    found = False
    for page in form.pages:
        if found:
            return page.attrib.get("id").strip("p")
        if current_page == int(page.attrib.get("id").strip("p")):
            # Return page id on next iteration.
            found = True
    else:
        # We did not found anything and the iteration ends.
        return current_page


def set_current_form_page(table, itemid, page, request):
    """Will save the given page of a form as the current page of the
    form in the session in the session. The key in the session is build
    as `<tablenameofitems>.<itemid>.form.page`

    :tabel: Name of the table where the items are stored.
    :itemid: ID of the item for which the page is stored.
    :page: ID of the page.
    :request: Current request
    :returns: Response
    """
    request.session['%s.%s.form.page' % (table, itemid)] = page
    request.session.save()


def get_item_from_request(request):
    """On every request pyramid will use a resource factory to load the
    requested resource for the current request. This resource is the
    context for the current request. This function will extract the
    loaded resource from the context. If the context is None, the item
    could not be loaded. In this case raise a 400 HTTP Status code
    exception.

    :request: Current request having the item loaded in the current context
    :returns: Loaded item

    """
    item = request.context.item
    if not item:
        raise HTTPBadRequest()
    return item



"""The term `Item` refers to an instance of a certain type of data which
is defined in of a :mod:`.modul`.  All items in Ringo are derived at
least from :class:`.BaseItem` and one ore more mixins from the
:mod:`.mixins` modul.

Ringo provides three classes to work with items in general.
:class:`.BaseList` is a class for handling listings of items. New items
are created by using a :class:`.BaseFactory`.
"""
import logging
import warnings
import operator
import math
import re
import uuid
import fuzzy
import Levenshtein
from sqlalchemy import Column, CHAR
from sqlalchemy.orm import joinedload, Session
from ringo.lib.helpers import (
    serialize, get_item_modul,
    get_raw_value, set_raw_value,
    prettify
)
from ringo.lib.cache import CACHE_MODULES
from ringo.lib.form import get_form_config
from ringo.lib.table import get_table_config
from ringo.lib.sql import DBSession
from ringo.lib.sql.cache import regions
from ringo.lib.sql.query import FromCache, set_relation_caching
from ringo.lib.alchemy import get_columns_from_instance
from ringo.model.mixins import StateMixin, Owned

log = logging.getLogger(__name__)


def levenshteinmatch(value, search, t):
    """Compares two strings with Levenshtein Algorithm and given
    threshold for similarity. It is a string metric for measuring the
    difference between two words, the minimum number of single-character
    edits. A threshold (between 0 and 1) can be applied for determining
    the maximum allowed single-character edits based on the word length
    of the longer word, potential floats are rounded up at this point."""
    lenV = len(value)
    lenS = len(search)

    if lenS >= lenV:
        l =  math.ceil(lenS * t)
    else:
        l = math.ceil(lenV * t)
    ld = Levenshtein.distance(value.encode("utf-8"), search.encode("utf-8"))
    return ld <= l


def doublemetaphone(value, search):
    """Compares two strings by applying the Double Metaphone phonetic
    encoding algorithm of the Fuzzy library using primary and secondary
    code of a string for matching."""
    dmeta = fuzzy.DMetaphone()
    dmeta_value = dmeta(value.encode("utf-8"))
    dmeta_search = dmeta(search.encode("utf-8"))

    if value == search:
        return True

    if dmeta_value[1] != None and dmeta_search[1] != None:
        for v in dmeta_value:
            for s in dmeta_search:
                if v == s:
                    return True 
    return dmeta_value == dmeta_search


def smatch(value, search):
    """Compares two strings by applying the Double Metaphone phonetic
    encoding algorithm and levenshteinmatch as second indicator in
    doubts.     """
    value = value.lower()
    search = search.lower()

    if value == search:
        return True

    if doublemetaphone(value, search):
        return True 
    else:
        return levenshteinmatch(search, value, 0.3) 


opmapping = {
    "<": operator.lt,
    "<=": operator.le,
    ">": operator.gt,
    ">=": operator.ge,
    "!=": operator.ne,
    "==": operator.eq,
    "~": smatch
}


 
def load_modul(item):
    """Will load the related modul for the given item. First we try to
    get the bound session from the object and reuse this session to load
    the modul item. If the item has no bound session then call the
    get_item_modul method with no request.

    :item: item
    :returns: modul instance

    """
    from ringo.model.modul import ModulItem
    session = Session.object_session(item)
    mid = item.__class__._modul_id
    # Loading the modul is expensive! So try to cache it.
    if not CACHE_MODULES.get(mid):
        if session:
            modul = session.query(ModulItem).filter_by(id=mid).one()
        else:
            modul = get_item_modul(None, item)
        CACHE_MODULES.set(modul.id, modul)
    return CACHE_MODULES.get(mid)


class BaseItem(object):
    """Base class for all items in Ringo. The class provides
    methods to get and set values of the instances.

    The class overwrites the `__getattr__` and `__setattr__` methods to
    support getting and setting dot separatedattributes like
    'getattr(foo.bar.baz)' 
    """

    _modul_id = None
    #  TODO: Check if its possible to set the modul of the class
    #  dynamically on application initialisation. This might help to get
    #  rid of the annoying loading of the modul for a class (ti)
    #  <2015-03-10 22:53>
    #__modul__ = None
    _sql_eager_loads = []
    """Configure a list of relations which are configured to be
    eager loaded."""

    uuid = Column('uuid', CHAR(36),
                  unique=True,
                  default=lambda x: str(uuid.uuid4()))

    def render(self):
        """This function can be used to render a different
        representation of the item. On default it also returns the
        simple string representation. Usefull to build some HTML used in
        links e.g overviews and lists"""
        return self.__str__()

    def __str__(self):
        return self.__unicode__()

    def __cmp__(self, other):
        """Comparision of the elements are done on their string
        representation"""
        s = unicode(self)
        o = unicode(other)
        if s < o:
            return -1
        elif s == o:
            return 0
        else:
            return 1

    def __getitem__(self, name):
        return self.get_value(name)

    def __getattr__(self, name):
        # In some cases it is needed to be able to trigger getting the
        # exapanded value without calling the get_value method. This can
        # be achieved by accessing the attribute with a special name.
        # Lets say you want to get the expanded value for `foo`. You get
        # this by asking for `foo__e_x_p_a_n_d`
        if name.endswith("__e_x_p_a_n_d"):
            return self.get_value(name.replace("__e_x_p_a_n_d", ""),
                                  expand=True)
        return get_raw_value(self, name)

    def __setattr__(self, name, value):
        return set_raw_value(self, name, value)

    def __unicode__(self):
        """Will return the string representation for the item based on a
        configured format string in the modul settings. If no format str
        can be found return the id of the item.

        The format string must have the following syntax:
        format_str|field1,field2,....fieldN. where format_str in a
        string with %s placeholders and fields is a comma separated list
        of fields of the item"""
        modul = load_modul(self)
        format_str, fields = modul.get_str_repr()
        if format_str:
            return format_str % tuple([prettify(None, self.get_value(f)) for f in fields])
        else:
            return "%s" % str(self.id or self.__class__)

    @classmethod
    def get_item_factory(cls, request=None):
        """Will return the :class:`.BaseFactory` for this class. Use
        this function to create and return specific factories for the
        class."""
        return BaseFactory(cls, request)

    @classmethod
    def _get_permissions(cls, modul, item, request):
        """Internal method to implement getting specific build of an ACL
        for this class (and instances). By default this function just
        calls the :func:`ringo.lib.security.get_permission` function.
        See this function for more details.

        If you need to implement an alternative permission check this
        function can be overwritten.  The `modul` and `item` attribute
        needed for the default get_permission method. `request` is used
        to get the current db session in case you need to reload other
        informations.

        :modul: Instance of the modul
        :item: Instance of the class
        :returns: List of permissions.

        """
        # FIXME: Try to get rid of the request element here. There
        # should not be any request needed in the model. This is needed
        # in the efa application for Participants to load additional
        # modules. (ti).
        # <2015-03-10 22:46>

        # FIXME: Circular import :((( Yes I know. We really need some
        # clean up here. (ti) <2015-03-10 21:08>
        from ringo.lib.security import get_permissions
        return get_permissions(modul, item)

    def reset_uuid(self):
        self.uuid = str(uuid.uuid4())

    def get_value(self, name, form_id="read", expand=False):
        """Return the value of the given attribute of the item. Unlike
        accessing the value directly this function this function
        optionally supports the expansion of the value before
        returning it. On default no expansion is done.

        Expansion is relevant for values which are saved as integer
        values in the database but have another literal meaning. This is
        typically true for values in selection lists. E.g the literal
        values for 'Yes' and 'No' are saves as value '1' and '0' in the
        database. To expand the value the method will try to get the
        literal value from the value in the database by looking in the
        form identified by the `form_id` attribute.

        :name: Name of the attribute with the value
        :form_id: ID of the form which will be used for expansion
        :expand: Expand the value before returning it
        :returns: Value of the named attribute
        """

        try:
            raw_value = getattr(self, name)
        except:
            # This error is only acceptable for blobforms as attributes
            # can be added and removed by the user on the fly. So there
            # is a good chance that older items do not have this attribute.
            if hasattr(self, 'id'):
                log.error("Attribute '%s' not found in '%s'; id:%s"
                          % (name, repr(self), self.id))
            else:
                log.error("Attribute '%s' not found in '%s'"
                          % (name, repr(self)))
            raw_value = None
        if expand:
            # In case the fieldname is dotted and refers to values in
            # related items then we need some special logic.
            elements = name.split(".")
            if len(elements) == 2:
                obj = getattr(self, elements[0])
                name = elements[1]
            elif len(elements) > 2:
                obj = getattr(self, ".".join(elements[0:-1]))
                name = elements[-1]
            else:
                obj = self

            # Expanding the value means to get the "literal" value for the
            # given value from the form.
            form_config = get_form_config(obj, form_id)
            try:
                field_config = form_config.get_field(name)
                options = []
                for option in field_config.options:
                    # Handle "list" values "{"",1,2}"
                    if str(raw_value).startswith("{") and str(raw_value).endswith("}"):
                        for value in raw_value.strip("}").strip("{").split(","):
                            if str(value) == str(option[1]):
                                options.append(option[0])
                    elif str(raw_value) == str(option[1]):
                        options.append(option[0])
                        break
                # If we can not match a value we return the raw value.
                # This can also happen if the user tries to expand value
                # which do not have options.
                if len(options) > 0:
                    return ", ".join(options)
                return raw_value
            except KeyError:
                # If the field/value which should to be expanded is not
                # included in the form the form library will raise a
                # KeyError exception. However this is not a big deal as
                # we still have the raw value and only the expandation
                # fails. So silently ignore this one. The exception is
                # already logged in the form library.
                pass
        return raw_value

    def get_values(self, include_relations=False, serialized=False):
        """Will return a dictionary with the values of the item. On
        default the function will return the pythonic values and
        excludes all related items. This behaviour can be changed by
        setting the `include_relations` and `serialized` option.

        In case the relations are included and the values are serialized
        then the serialized value of a related item is its id.

        :include_relations: Flag if relations should be included in the
                            returned dictionary.
        :serialized: Flag if the values should be serialized.
        :returns: Dictionary with key value pairs.
        """
        values = {}
        for field in get_columns_from_instance(self, include_relations):
            # Ignore private form fields
            if field.startswith("_"):
                continue
            if serialized:
                value = serialize(getattr(self, field))
            else:
                value = getattr(self, field)
            values[field] = value
        return values

    def set_values(self, values):
        """Will set the values of the item. The values to be set are
        provided by a dictionary with key value pairs given with the
        `values` option. Keys in the dictionary beginning with "_" are
        considered private and are ignored.

        This function does not explicit handle saving the changed data!
        Saving data (E.g making a new instance of an item persistent in
        the DB) is handled by the :func:`.BaseItkkem.save` method.

        .. note::
            Setting the values for foreign key attributes might not work
            as expected. This is especially true for foreign keys to
            already existing items. You are not able to change a
            existing relation in the items by changing the foreign key
            value (You must do this by setting the related item). In
            this case the value for the foreign key seems to be ignored
            and replaced by the one of the actual related item.  In
            contrast you can set a new relation by setting the foreign
            key if this is a new relation.

        :values: Dictionary with values to be set
        """

        for key, value in values.iteritems():
            # Ignore private form fields
            if key.startswith('_'):
                continue
            if hasattr(self, key):
                log.debug("Setting value '%s' in %s" % (value, key))
                oldvalue = getattr(self, key)
                if isinstance(value, list) and isinstance(oldvalue, list):
                    # Special handling for relations in NM relations.
                    # See ticket #19 in Ringo issue tracker for more
                    # details. Simply exchaning the relations in this
                    # case seems not to work. I case of the error in
                    # triggering the deleting of items in the
                    # nm-relation from both sides. In this case the
                    # second deletion will fail.
                    #
                    # Regarding to
                    # http://docs.sqlalchemy.org/en/latest/orm/basic_relationships.html#deleting-rows-from-the-many-to-many-table
                    # relations should be removed by using remove(item).
                    # Howver poping the items before adding the new
                    # items seems to work too. I excpet that poping the
                    # items will somehow tweak SQLAlchemy in the way
                    # that it handles deleting items correct now.
                    while oldvalue:
                        oldvalue.pop(0)
                    for nvalue in value:
                        oldvalue.append(nvalue)
                else:
                    setattr(self, key, value)
            else:
                log.warning('Not saving "%s". Attribute not found' % key)

    def save(self, data, request=None):
        """Method to set new values and 'saving' changes to the item. In
        contrast to just setting the values saving means triggering
        additional actions like handling state changes.

        The function will call the :func:`.BaseItem.set_values` function
        to set the values so setting the values in a separate call of
        this method is not needed. Actually doing this can cause
        problems as this function can not determine changes between old
        and new values properly anymore.

        If the current item has no value for the id attribute it is
        assumed that this item must be added to the database as a new
        item. In this case you need to provide a dbsession (as part of
        the request) as the new item is not linked to any dbsession yet.

        .. note::
            Making the changes persistent to the database is finally
            done on the end of the request when the transaction of the
            current session is commited. 

        Please note, that you must ensure that the submitted values are
        validated. This function does no validation on the submitted
        data.

        :data: Dictionary with key value pairs.
        :request: Current request session. Used when saving new items.
        :returns: item with new data.

        """
        if request:
            dbsession = request.db
        else:
            dbsession = None
        # Use repr here a the __unicode__ method default to self.id
        # which might not be available here
        log.debug("Saving %s" % repr(self))

        old_values = self.get_values()
        # Set values
        self.set_values(data)

        # Handle statechange
        if isinstance(self, StateMixin):
            for key in self.list_statemachines():
                new_state_id = data.get(key)
                old_state_id = old_values.get(key)
                if ((new_state_id and old_state_id)
                   and (new_state_id != old_state_id)):
                    self.change_state(request, key,
                                      old_state_id, new_state_id)

        #  FIXME: Fix building a log entry. The log modul has been
        #  converted to an extension and is not available in the ringo
        #  core (ti) <2015-01-31 19:04>
        ## Handle logentry
        #if isinstance(self, Logged):
        #    if not self.id:
        #        subject = "Create"
        #        text = json.dumps(self.build_changes(old_values, data))
        #    else:
        #        subject = "Update"
        #        text = json.dumps(self.build_changes(old_values, data))
        #    self.add_log_entry(subject, text, request)

        # If the item has no id, then we assume it is a new item. So
        # add it to the database session.
        if not self.id and dbsession:
            dbsession.add(self)
            # flush the session to make the new id in the element
            # presistent.
            dbsession.flush()

            # Check if uid or gid or uid inheritance is configured.
            if isinstance(self, Owned):
                gid_relation = getattr(self, '_inherit_gid')
                if gid_relation:
                    parent = getattr(self, gid_relation)
                    # FIXME: Check why this attribute can be None. (ti)
                    # <2014-04-01 13:37>
                    if parent:
                        self.group = parent.group
                    else:
                        log.warning("Inheritance of group '%s' failed. "
                                    "Was None" % gid_relation)
                uid_relation = getattr(self, '_inherit_uid')
                if uid_relation:
                    parent = getattr(self, uid_relation)
                    if parent:
                        self.owner = parent.owner
                    else:
                        log.warning("Inheritance of group '%s' failed. "
                                    "Was None" % gid_relation)
        return self

########################################################################
#                               BaseList                               #
########################################################################


def get_item_list(request, clazz, user=None, cache="", items=None):
    """Returns a :class:`.BaseList` instance with items of the given
    clazz. You can optionally provide a user object. If provided the
    list will only contain items which are readable by user in the
    current request.  Further you can define a caching region to cache
    the results of the sqlquery. If not provided no caching is done.

    :request: Current request
    :clazz: Clazz for with the items in the baselist will be loaded.
    :user: If provided only items readable for the
           user in the current request are included in the list
    :cache: Name of the cache region. If empty then no caching is
            done.
    :items: Set items of the Baselist. If provided no items will be
    loaded.
    :returns: BaseList instance

    """
    if user:
        user_key = user.id
    else:
        user_key = None
    key = "%s-%s" % (clazz._modul_id, user_key)
    if not request.cache_item_list.get(key):
        listing = BaseList(clazz, request.db, cache, items)
        if user:
            listing = filter_itemlist_for_user(request, listing)
        # Only cache the result if not items has been prefined for the
        # list.
        if items is None:
            request.cache_item_list.set(key, listing)
            return listing
        else:
            return listing
    return request.cache_item_list.get(key)


def filter_itemlist_for_user(request, baselist):
    """Returns a filterd baselist. The items are filterd based on
    the permission of the current user in the request. Only items are
    included where the given user has read access. This means:

     1. The user has aproriate role (read access)
     2. Is owner or member of the items group

    :request: Current request
    :baselist: BaseList instance
    :returns: Filtered BaseList instance

    """
    from ringo.lib.security import has_permission
    filtered_items = []
    if request.user and not request.user.has_role("admin"):
        # Iterate over all items and check if the user has generally
        # access to the item.
        for item in baselist.items:
            # Only check ownership if the item provides a uid.
            if has_permission('read', item, request):
                filtered_items.append(item)
        baselist.items = filtered_items
    return baselist


class BaseList(object):
    """Base class for listing of items in Ringo. The class provides
    methods for sorting and filtering the items of the list.

    The BaseList is usually created by calling the
    :func:`.get_item_list` method. Using this method the BaseList will
    include all items of a given class which are readable by the current
    user. 

    Alternatively the BaseList can be initiated directly in two ways. On
    default all items of a given class will be loaded from the database.
    The other way is to initiate the list with a list of preloaded
    items.
    """
    def __init__(self, clazz, db, cache="", items=None):
        """A List object of. A list can be filterd, and sorted.

        :clazz: Class of items which will be loaded.
        :db: DB connection used to load the items.
        :cache: Name of the cache region. If empty then no caching is
                done.
        :items: Set items of the Baselist. If provided no items will be
        loaded.
        """
        self.clazz = clazz
        self.db = db
        if items is None:
            q = self.db.query(self.clazz)
            if cache in regions.keys():
                q = set_relation_caching(q, self.clazz, cache)
                q = q.options(FromCache(cache))
            for relation in self.clazz._sql_eager_loads:
                q = q.options(joinedload(relation))
            self.items = q.all()
        else:
            self.items = items
        self.search_filter = []

    def __iter__(self):
        return iter(self.items)

    def sort(self, field, order, expand=False):
        """Will return a sorted item list. Sorting is done based on the
        string version of the value in the sort field.

        :field: Name of the field on which the sort will be done
        :order: If "desc" then the order will be reverted.
        :expand: If True, then the sorting will be done on the expanded values.
        :returns: Sorted item list

        """
        def attrgetter(field, expand):
            def g(obj):
                return obj.get_value(field, expand=expand)
            return g

        def nonecmp(a, b):
            if a is None and b is None:
                return 0
            if a is None:
                return -1
            if b is None:
                return 1
            return cmp(a, b)

        sorted_items = sorted(self.items,
                              cmp=nonecmp,
                              key=attrgetter(field, expand))
        if order == "desc":
            sorted_items.reverse()
        self.items = sorted_items

    def paginate(self, page=0, size=None):
        """This function will set some internal values for the
        pagination function based on the given params.

        :page: Integer of the current page
        :size: Items per page
        :returns:

        """
        self.pagination_size = size
        """Number of items per page"""
        self.pagination_current = page
        """Current selected page"""
        self.pagination_pages = None
        """Number of total pages in the pagination"""
        self.pagination_first = None
        """First page in pagination navigation"""
        self.pagination_last = None
        """Last page in pagination navigation"""
        self.pagination_start = None
        """Start index for slicing the items list on the current page"""
        self.pagination_end = None
        """End index for slicing the items list on the current page"""

        # Calulate the slicing indexes for paginating and the total
        # number of pages.
        if size is None:
            pages = 1
            self.pagination_start = 0
            self.pagination_end = len(self.items)
        else:
            pages = int(math.ceil(float(len(self.items)) / size))
            self.pagination_start = page * size
            self.pagination_end = (page + 1) * size
        self.pagination_pages = pages

        ## Calculate the range to the pagination navigation
        start = self.pagination_current - 4
        end = self.pagination_current + 5
        if start < 0:
            end += abs(start)
            start = 0
        elif end > self.pagination_pages:
            end = self.pagination_pages
            start -= (abs((self.pagination_pages + 5) - end))
        self.pagination_first = start
        self.pagination_last = end

    def filter(self, filter_stack):
        """This function will filter the items by only leaving
        those items in the list which match all search criteria in the
        filter stack. The number of items will get reduced with every
        iteration on the filter stack. The search is case sensitive.

        The filter stack is a list of tuples which describe the filter
        expression. Each tuple consists of three values:

        1. The search expression
        2. Optionally the name of the column on which the search will be
           applied.
        3. Boolean flag to indicate that the given search expression
           should be handled as a regular expression.

        The search support three modes. On default the search will look
        for the presence of the given search string within the values of
        the item. It will also match on parts of the string. So
        searching for 'Foo' will also match 'Foobar'. Another mode is
        using the search expression as a regular expression. The last
        mode makes use of special operators. The search supports the
        following operators: "<", "<=", "!=", ">" ">=" and "~" The
        operator can be provided with the search string.  It mus be the
        first word of the search expression. If a operator is present it
        will be used.

        The "~" operator will trigger a fuzzy search using the Double
        Metaphone algorithm for determining equal phonetics. If the
        phonetics do not match the Levenshtein distance will be
        calculated. 

        The search will iterate over all items in the list. For each
        item the function will try to match the value of either all, or
        from the configured search field with the regular expression or
        configured operator. If the value matches, then the item is kept
        in the list.

        :filter_stack: Filter stack
        :returns: Filtered list of items
        """
        self.search_filter = filter_stack
        log.debug('Length filterstack: %s' % len(filter_stack))
        table_config = get_table_config(self.clazz)
        table_columns = {}

        # Save cols in the tableconfig for later access while getting values.
        for col in table_config.get_columns():
            table_columns[col.get('name')] = col

        for search, search_field, regexpr in filter_stack:
            search_op = None
            # Get soperator
            x = search.split(" ")
            if x[0] in opmapping:
                search_op = x[0]
                search = " ".join(x[1:])
            # Build a regular expression
            if regexpr:
                re_expr = re.compile(search, re.IGNORECASE)
            else:
                re_expr = re.compile(re.escape(search), re.IGNORECASE)
            filtered_items = []
            log.debug('Filtering "%s" in "%s" with operator "%s" on %s items'
                      % (search, search_field, search_op, len(self.items)))
            if search_field != "":
                fields = [search_field]
            else:
                fields = table_columns.keys()
            for item in self.items:
                for field in fields:
                    expand = table_columns[field].get('expand')
                    value = item.get_value(field, expand=expand)
                    if isinstance(value, list):
                        value = ", ".join([unicode(x) for x in value])
                    else:
                        value = unicode(value)
                    if search_op:
                        if opmapping[search_op](value, search):
                            filtered_items.append(item)
                            break
                    else:
                        if re_expr.search(value):
                            filtered_items.append(item)
                            break
            self.items = filtered_items


class BaseFactory(object):
    """Factory class to create new instances of :class:`.BaseItem` or
    derived classes. Usually the factory for items of a certain module
    can be initiated by calling the :func:`.BaseItem.get_item_factory`
    class method."""

    def __init__(self, clazz, request=None):
        """Inits the factory.

        :clazz: The clazz of which new items will be created

        """
        self._clazz = clazz
        self._request = request

    def create(self, user, values):
        """Will create a new instance of clazz. The instance is it is
        not saved persistent at this moment. The method will also take
        care of setting the correct ownership.

        :user: User instance will own the new created item
        :values: Dictionary with values for the new item
        :returns: Instance of clazz

        """
        if not isinstance(values, dict):
            raise ValueError("Values must be a dictionary")
        item = self._clazz()
        # Try to set the ownership of the entry if the item provides the
        # fields.
        if (hasattr(item, 'uid')
           and user is not None):
            item.uid = user.id
        if (hasattr(item, 'gid')):
            modul = get_item_modul(None, item)
            if modul.default_gid:
                item.gid = modul.default_gid
            elif (user is not None and user.default_gid):
                item.gid = user.default_gid
        if values:
            item.set_values(values)
        return item

    def load(self, id, db=None, cache="", uuid=False, field=None):
        """Loads the item with id from the database and returns it.

        :id: ID of the item to be loaded
        :db: DB session to load the item
        :cache: Name of the cache region. If empty then no caching is
                done.
        :uuid: If true the given id is a uuid. Default to false
        :field: If given the item can be loaded by an alternative field.
                Default to None
        :returns: Instance of clazz

        """
        if db is None:
            db = DBSession
        q = db.query(self._clazz)
        if cache in regions.keys():
            q = set_relation_caching(q, self._clazz, cache)
            q = q.options(FromCache(cache))
        for relation in self._clazz._sql_eager_loads:
            q = q.options(joinedload(relation))
        if field:
            return q.filter(getattr(self._clazz, field) == id).one()
        if uuid:
            warnings.warn("Use of 'uuid' is deprecated in load function "
                          "and will be removed in the future. Use "
                          "field='uuid' instead", DeprecationWarning)
            return q.filter(self._clazz.uuid == id).one()
        else:
            return q.filter(self._clazz.id == id).one()



import logging
import hashlib
import uuid
import string
import random
from passlib.context import CryptContext
from datetime import datetime
from pyramid.events import ContextFound, NewRequest
from pyramid.security import unauthenticated_userid, \
    has_permission as has_permission_, \
    Allow, ALL_PERMISSIONS
from pyramid.authentication import AuthTktAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.httpexceptions import HTTPUnauthorized
from sqlalchemy.orm.exc import NoResultFound
from ringo.lib.helpers import get_item_modul
from ringo.lib.sql import DBSession
from ringo.lib.alchemy import get_relations_from_clazz
from ringo.model.base import BaseItem
from ringo.model.modul import ModulItem
from ringo.model.user import User, PasswordResetRequest, Login

log = logging.getLogger(__name__)


def password_generator(size=12, chars=string.ascii_uppercase + string.digits):
    return ''.join(random.choice(chars) for x in range(size))


def get_auth_timeout(settings):
    """Will return the amount of seconds until the auth session will
    time out. This can be configured in the application ini file. If no
    configuration is found. it defaults to 1800 seconds (30min)"""
    return int(settings.get("auth.timeout") or 1800)


def get_auth_timeout_warning(settings):
    """Will return the amount of seconds before a warning dialog will be
    raised before the session will time out. This can be configured in
    the application ini file. If no configuration is found. it defaults
    to 30 seconds"""
    return int(settings.get("auth.timeout_warning") or 30)


def get_cookie_secret(settings):
    """Will return the configured string in the config to sign the
    cookies. If no string is configured. Generate a random string for
    this."""
    secret = settings.get("security.cookie_secret")
    if not secret:
        secret = password_generator(50)
    return secret


pwd_context = CryptContext(
    # replace this list with the hash(es) you wish to support.
    # this example sets pbkdf2_sha256 as the default,
    # with support for legacy md5 hashes.
    schemes=["pbkdf2_sha256"],
    default="pbkdf2_sha256",

    # vary rounds parameter randomly when creating new hashes...
    all__vary_rounds=0.1,

    # set the number of rounds that should be used...
    # (appropriate values may vary for different schemes,
    # and the amount of time you wish it to take)
    pbkdf2_sha256__default_rounds=8000,
)


def encrypt_password(password, scheme=None):
    """Will return a string with encrypted password using the passlib
    library. The returned string will have the following format:
    $algo$rounds$salt$hash

    :password: unencrypted password
    :scheme: string of an supported encryption algorithm. Defaults tthe
    CryptContext default
    :returns: encrypted password

    """
    return pwd_context.encrypt(password, scheme=scheme)


def verify_password(password, pwhash):
    """Will return True if the password is valid. Else False.

    :password: unencrypted password
    :pwhash: encrypted password
    :returns: True or False
    """
    if pwhash.startswith("$"):
        return pwd_context.verify(password, pwhash)
    else:
        # Old password hash scheme without salt etc.
        md5_pw = hashlib.md5()
        md5_pw.update(password)
        return pwhash == md5_pw.hexdigest()


def passwords_needs_update(pwhash):
    """Will return True if the given pwhash uses a deprecated
    encryption algorithm and needs to be updated.

    :pwhash: encrypted password
    :returns: True or False
    """
    if pwhash.startswith("$"):
        return pwd_context.needs_update(pwhash)
    else:
        # Old password hash scheme without salt etc.
        return True


def load_user(login):
    """Will return the user with the given login
    name. If no user can be found with the given login "None" will be
    returned.

    :login: Login of the user as string
    :returns: User or None

    """
    try:
        return DBSession.query(User).filter_by(login=login).one()
    except NoResultFound:
        return None


def csrf_token_validation(event):
    request = event.request
    if (request.method == 'POST'):
        csrf = request.params.get('csrf_token')
        if (csrf != unicode(request.session.get_csrf_token())):
            log.warning("CSRF token check failed. Raising 401 exception.")
            raise HTTPUnauthorized


def refresh_auth_cookie(event):
    """This fuction will refresh the the timeout of the authentification
    cookie. See
    https://groups.google.com/forum/#!topic/pylons-discuss/HczvAEd5xY8
    for more details.
    """
    event.request.unauthenticated_userid


def setup_ringo_security(config):
    settings = config.registry.settings
    timeout = get_auth_timeout(settings) + 5
    secret = get_cookie_secret(settings)
    secure = settings.get("security.cookie_secure", "false") == "true"
    include_ip = settings.get("security.cookie_ip", "false") == "true"
    path = settings.get("security.cookie_path", "/")
    domain = settings.get("security.cookie_domain")
    httponly = settings.get("security.cookie_httponly", "false") == "true"
    cookie_name = settings.get("security.cookie_name", "auth_tkt")
    authn_policy = AuthTktAuthenticationPolicy(secret,
                                               secure=secure,
                                               hashalg='sha512',
                                               timeout=timeout,
                                               reissue_time=timeout / 10,
                                               callback=get_principals,
                                               include_ip=include_ip,
                                               path=path,
                                               domain=domain,
                                               http_only=httponly,
                                               cookie_name=cookie_name)
    authz_policy = ACLAuthorizationPolicy()
    config.set_authorization_policy(authz_policy)
    config.set_authentication_policy(authn_policy)
    # Make the user object available as attribute "user" in the request.
    # See http://docs.pylonsproject.org/projects/ \
    #            pyramid_cookbook/en/latest/auth/ \
    #            user_object.html
    config.add_request_method(get_user, 'user', reify=True)

    # Add subscriber to check the CSRF token in POST requests. You can
    # disable this for testing by setting the
    # "security.enable_csrf_check" config variable to "false".
    if settings.get('security.enable_csrf_check', 'true') != "false":
        config.add_subscriber(csrf_token_validation, ContextFound)
    # Refresh the auth cookie timeout on every request. On default this
    # would only happen on requests which needs
    # authentification/authorisation. As the authentification should be
    # valid as long the user shows some activity by triggering requests
    # this tween will refresh the timeout on every request.
    config.add_subscriber(refresh_auth_cookie, NewRequest)

    # Add tweens to add custom security headers.
    # http://ghaandeeonit.tumblr.com/post/65698553805/securing-your-pyramid-application
    if settings.get("security.header_secure", "true") == "true":
        config.add_tween('ringo.tweens.secure_headers.secure_headers_factory')
    if settings.get("security.header_clickjacking", "true") == "true":
        config.add_tween('ringo.tweens.clickjacking.clickjacking_factory')
    if settings.get("security.header_csp", "false") == "true":
        config.add_tween('ringo.tweens.csp.csp_factory')


def get_user(request):
    userid = unauthenticated_userid(request)
    if userid is not None:
        return _load_user(userid, request)
    return None


def has_permission(permission, context, request):
    """Wrapper for pyramid's buitin has_permission function.  This
    wrapper sets dynamically the __acl__ attribute of the given context
    and then  .  check if the user has the given permission in the
    current context using pyramid's has_permission function.

    Context can be:
    * Instance of BaseItem
    * Subclass of BaseItem
    * Ressource, built from a RessourceFactory

    If context is an instance or subclass of BaseItem the wrapper will
    dynamically set the __acl__ attribute. This attribute is used by the
    pyramid's has_permission function the check the permission. If the
    context is a resource the function does nothing as the resource
    already has the __acl__ attribute set.

    If the user has the permission the it returns True, else False
    (Actually it returns a boolean like object, see pyramids
    has_permission doc for more details.)

    :permission: String. Name of the permission. E.g list, create, read
    :context: Either Resource, Instance of BaseItem or Subclass of BaseItem
    :request: current request
    :returns: True or False (Boolean like object)

    """
    if isinstance(context, BaseItem) or hasattr(context, "_modul_id"):
        modul = get_item_modul(request, context)
        context.__acl__ = context._get_permissions(modul, context, request)
    # Call of has_permission will trigger 4 additional SQL-Queries. The
    # query will only be trigger once per request.
    return has_permission_(permission, context, request)


def get_permissions(modul, item=None):
    """Will return a list permissions attached to the modul and
    optionally to particular item of the modul. The returned list is
    suitable for using it as ACL in pyramid's authorization system.  The
    function will at least return class based permissions. If a item is
    provided then additional item level permissions are returned.

    Note that item can also be a class and not an instance. In this case
    the class level permissions are returned

    The function will iterate over all available actions of the given
    modul and tries to adds a permission entry to the for every role
    which has granted access to the action.

    For every role it will check the following things:

    1. If the role is marked as "administrational" role, then add the
       permission for role and action.
    2. Elif action is either "list" or "create" (permissons on class
       level) then add the permission for the role and action.
    3. Elif an item is provided add item level permissions for the role
       _and_ only the owner or users which are member of the items group.

    :model: The modul for which the permissions are returned
    :item: Optional: Item of the model for which the permissons as
           returned
    :returns: List of permissions
    """
    perms = []
    # Default permisson. Admins should be allowed to to everything.
    perms.append((Allow, 'role:admin', ALL_PERMISSIONS))
    if not modul:
        return perms

    # Load current states for the item as we need to check which
    # permissions the user has depending on the current states of the item.
    current_states = []
    if isinstance(item, BaseItem) and hasattr(item, '_statemachines'):
        for smname in item._statemachines:
            sm = item.get_statemachine(smname)
            state = sm.get_state()
            current_states.append(state)

    # No need to call get_item_actions. We only need the modul actions
    # here as all other ActionItem added dynamically to the items clazz
    # (e.g Mixin actions), will map their permission to one of the
    # available base modul permissions (create, read, update, delete...)
    actions = modul.actions
    for action in actions:
        permission = action.permission or action.name.lower()
        # TODO: Note that actions, which are not inserted into the
        # database (e.g custom mixin actions.) will not have any roles
        # and are not considerd on building principals yet. (ti)
        # <2014-02-07 11:08>
        for role in action.roles:
            add_perm = True
            default_principal = 'role:%s' % role.name

            # Check if the actions is available in the current state of
            # the item if the item has states.
            for state in current_states:
                if str(action.name.lower()) in state.get_disabled_actions(role.name):
                    add_perm = False
                    continue

            # administrational role means allow without further
            # ownership checks.
            if role.admin is True and add_perm:
                perms.append((Allow, default_principal, permission))

            # Modul level (class level) permissions.
            # Always add the default principals for the create and list
            # actions as those actions can not be checked on item level
            # anyway.
            elif permission in ['create', 'list']:
                perms.append((Allow, default_principal, permission))
            # If the item has a uuid the we want get the permission on
            # Item level. Only allow the owner or members of the items
            # group.
            elif item and hasattr(item, 'uid') and add_perm:
                principal = default_principal + ';uid:%s' % item.uid
                perms.append((Allow, principal, permission))
                principal = default_principal + ';group:%s' % item.gid
                perms.append((Allow, principal, permission))
            # Finally the item is not an instance of a BaseItem then add
            # remaining action we want to get the permission on modul
            # level too. So again add the default principal
            elif not isinstance(item, BaseItem) and add_perm:
                perms.append((Allow, default_principal, permission))
    return perms


def __add_principal(principals, new):
    if new not in principals:
        principals.append(new)
    return principals


def get_principals(userid, request):
    """Returns a list of pricipals for the user with userid for the
    given request.

    Principals are basically strings naming the groups or roles the user have.
    Example: role:admin or group:users are typical principals.

    :userid: id of the user
    :request: current request
    :returns: list with pricipals

    """
    if request.user:
        user = request.user
    else:
        user = _load_user(userid, request)
    principals = []
    if user:
        # Add the roles of the user. The roles will be added in
        # connection with the uid and groups as the roles should only be
        # applicable if the user is the owner or member of the group of
        # the item.
        for urole in user.roles:
            principal = 'role:%s' % urole.name
            __add_principal(principals, principal)
            principal = 'role:%s;uid:%s' % (urole.name, user.id)
            __add_principal(principals, principal)
            for group in user.groups:
                # Add the user role for every group the user is member
                # of
                principal = 'role:%s;group:%s' % (urole.name, group.id)
                __add_principal(principals, principal)
        # Finally add the user itself
        principal = 'uid:%s' % user.id
        __add_principal(principals, principal)
    principals.sort(key=len, reverse=True)
    log.debug('Principals for userid "%s": %s' % (userid, principals))
    return principals


# ROLES
#######


def has_role(user, role):
    """Return True if the user has the given role. Else False"
    :user: User instance
    :returns: True or False
    """
    if user is None:
        return False
    return user.has_role(role)


def get_roles(user):
    """Returns a list of roles the user has. The list contains
    `Role` object and are collected by loading roles directly
    attached to the user plus roles attached to the groups the user
    is member of

    :user: User instance
    :returns: List of `Role` instances

    """
    return user.roles


# GROUPS
########


def has_group(user, group):
    """Return True if the user is in the the given group. Else False"
    :user: User instance
    :returns: True or False
    """

    groups = [g.name for g in user.groups]
    return group in groups


# Helpers
#########


def _load_user(userid, request):
    try:
        factory = User.get_item_factory()
        return factory.load(userid)
    except NoResultFound:
        return None


def activate_user(token, db):
    """Will check if confirmation token is valid by searching a user
    with this activation token. If a user is found the activated flag
    will be set to true.

    :token: confirmation token
    :db: db connection
    :returns: user
    """
    try:
        user = db.query(User).filter_by(activation_token=token).one()
        user.activated = True
        user.activation_token = ""
        log.info('User %s activated' % user)
        return user
    except NoResultFound:
        log.warning('User activation failed for token %s'
                    % (token))
        return None


def password_reset(token, db):
    """Will check if there is a password reset token is valid. The
    function will reset the password of the user. The new generated
    password is returned.

    :token: password reset token
    :db: db connection
    :returns: tupe of user and password
    """
    try:
        token = db.query(PasswordResetRequest).filter_by(token=token).one()
        # Check that the token is not outdated
        td = datetime.now() - token.created
        if td.days <= 1:
            user = token.user
            password = password_generator()
            user.password = encrypt_password(password)
            log.info('Password reset success for user %s' % user)
            # delete all old password request token
            for old_token in user.reset_tokens:
                db.delete(old_token)
            return user, password
        else:
            log.warning('Password reset failed for token %s (outdated)'
                        % token)
            return None, None
    except NoResultFound:
        log.warning('Password reset failed for token %s' % token)
        return None, None


def request_password_reset(username, db):
    """Will check if there is a user with the given username and creates
    a new entry in the password_reset table with a new token. The new
    token is than returned. If there is no user with the username then
    do nothing and return None.

    :username: username
    :db: db connection
    :returns: User

    """
    user = None
    try:
        user = db.query(User).filter_by(login=username).one()
        log.info('Password reset request for user %s' % user)
        reset_token = PasswordResetRequest(str(uuid.uuid4()))
        user.reset_tokens.append(reset_token)
    except NoResultFound:
        log.warning('Password reset request for non existing user %s'
                    % username)
    return user


def login(username, password):
    """Returns a `User` instance if the login does not fail with the
    given login and password.

    :username: username as String
    :password: password as SHA1 encrypted string
    :returns: `User` instance if login is OK else None.

    """
    log.debug("Login user '%s' with pw '%s'" % (username, password))
    user = load_user(username)
    if user:
        if verify_password(password, user.password):
            if user.activated:
                Login(user, success=True)
                user.last_login = datetime.utcnow()
                log.info("Login successfull '%s'" % (username))
                if passwords_needs_update(user.password):
                    log.info("Updating password for user '%s'" % (username))
                    user.password = encrypt_password(password)
                return user
            else:
                Login(user, success=False)
                log.info("Login failed for user '%s'. "
                         "Reason: Not activated" % username)
                return user
        else:
            Login(user, success=False)
            log.info("Login failed for user '%s'. "
                     "Reason: Wrong password" % username)
    else:
        log.info("Login failed for user '%s'. "
                 "Reason: Username not known" % username)
    return None


def get_last_successfull_login(request, user):
    if user is not None:
        result = request.db.query(Login) \
            .filter(Login.success == True, Login.uid == user.id) \
            .order_by(Login.datetime.desc()) \
            .limit(2).all()
        if len(result) > 1:
            return result[1]
    return None


def get_last_failed_login(request, user):
    result = None
    if user is not None:
        result = request.db.query(Login) \
            .filter(Login.success == False, Login.uid == user.id) \
            .order_by(Login.datetime.desc()) \
            .first()
    return result


def get_last_logins(request, since, success=None):
    """Returns a list of last logins since the given datetime.

    :request: Current request
    :since: Datetime object
    :success: Boolean flag to indicate to only return successfull or
              failed logins. If None all logins are returned.
    :returns: List of Login items

    """
    if request.user is None:
        return []

    result = request.db.query(Login) \
        .filter(Login.datetime > since, Login.uid == request.user.id) \
        .order_by(Login.id.desc()) \
        .all()
    if success is None:
        return result
    else:
        return [l for l in result if l.success == success]


class AuthentificationException(Exception):
    """Exception to be raise if error on authentification is detected."""
    pass


class AuthorizationException(Exception):
    """Exception to be raise if a authorization error is detected."""
    pass


class ValueChecker(object):
    """Checks the permission of a user to set values, provided by an
    dictionary, in an item.

    This checker restricts users to only set the values which they may
    also read. This prevents the possible security thread
    of gaining higher privileges by setting forbidden values with a
    manipulated requests.

    This espacially important for values which will change relations in
    the item. For each relation of the item the function will iterate
    over the values in the dictionary and check if the user is at least
    allowed to read the value.

    If the user is not allowed to read a value the checker will either

     * Raise an AuthorizationException if strict mode is enabled, or
     * Filter the set of values based on the permissions.

    In all cases a warning is logged if the Checker detects a permission
    violation. This class implements a very important security check and
    should be always used before values are about to be changed in an
    item.

    Limitations:
    The function currently only checks permissions on values which are
    about to be set in a sqlalchemy.orm.relation. This means setting the
    values directly in a FK is still possible.
    """

    def __init__(self, strict=True):
        self.strict = strict
        """Mode of the checker. If strict is true a
        AuthorizationException will be raised if a permission violation
        is detected. Otherwise the values get filtered."""

    def check(self, clazz, values, request, item=None):
        """Will do permission checks on the values. If strict mode is
        disabled the function will do the following:

         * For new added relations the fuction will remove every
           relation from to the values for which the user is not granted
           read access.
         * For removed relations the function will re-add every relation
           to the values for which the user is not granted read access.

        The function will return the values after checking and
        filtering. If strict mode is enabled an AuthorizationException
        is raised.

        :clazz: The clazz of the item which is about to be changed
        :values: Dictionary with deserialised and validated form values.
        :request: Current request
        :item: Defaults to None. If provided than the checker will only
        check differences between the values to be set and already
        present in the item. Else all values are checked.
        """
        for relation in get_relations_from_clazz(clazz):
            # If the relation is not set in the values then continue, as
            # we do not need to check anything. 
            if relation not in values:
                continue

            # Get old and new values.
            # For the purpose to unify the test logic of the permissions
            # checks the values will be converted into a list in all
            # cases. In case of N:1 realtions new_values will be a
            # single item and not a list.  In all other cases (1:N, N:N)
            # new_values is a list.
            new_values = values[relation]
            if not isinstance(new_values, list):
                new_values = [new_values]

            if item is not None:
                old_values = item.get_value(relation)
                if not isinstance(old_values, list):
                    old_values = [old_values]
            else:
                old_values = []

            # Determine the values which actually need to be checked.
            # If no item is provided, all values are checked. Otherwise
            # only changed values are checked.
            if item is None:
                to_check = [(v, 0) for v in new_values]
            else:
                to_check = self._diff(old_values, new_values)

            # Now iterate over all value which need to be checked.
            for value, modifier in to_check:
                # If the relation is a ModulItem or is not a Baseitem at
                # all also do no checks.
                # Modulitem do not have any uid or gid which will allow
                # checking permissions. Allowing links to a modul item
                # is currently not known to be a security thread.
                if (isinstance(value, ModulItem)
                    or has_permission("read", value, request)
                    or not isinstance(value, BaseItem)):
                    continue
                else:
                    if modifier > 0:
                        action = "add"
                    elif modifier < 0:
                        action = "remove"
                    else:
                        action = "set"
                    msg = ("User '%s' is not allowed to %s %s %s from %s %s"
                           % (request.user.login, action, type(value),
                              value.id, clazz, getattr(item, "id", '')))
                    log.warning(msg)
                    if self.strict:
                        raise AuthorizationException(msg)
                    else:
                        raise NotImplementedError()
        return values

    def _diff(self, old, new):
        """Will diff the values of old and new. It returns a list of
        tuples with values

          * Which are exclusive in new (value has been added)
          * Which are exclusive in old (value has been removed)

        The tuple will contain the value and 1 for added values and -1
        for removed items.
        """
        diff = []
        if isinstance(old, list):
            old = set(old)
            new = set(new)
            # Get added
            for added in new.difference(old):
                diff.append((added, 1))
            # Get removed
            for removed in old.difference(new):
                diff.append((removed, -1))
            return diff
        else:
            if old != new:
                if old is not None:
                    diff.append((old, -1))
                if new is not None:
                    diff.append((new, 1))
            return diff



"""Modul with various methods to format and transform values"""
from datetime import (
    datetime,
    date,
    timedelta
)
from dateutil import tz
from babel.core import Locale
from babel.dates import (
    format_datetime as babel_format_datetime,
    format_date as babel_format_date
)
from webhelpers.html import literal
from pyramid.threadlocal import get_current_request
from pyramid.i18n import get_locale_name
from formbar.converters import from_timedelta

########################################################################
#                          Formating content                           #
########################################################################


def prettify(request, value):
    """Generic function used in template rendering to prettify a given
    pythonic value into to a more human readable form. Depending on the
    datatype the function will call a specialised formatting function
    which takes care of localisation etc.

    The locale is determined from the given request.

    :request: Current request
    :value: Pythonic value
    :returns: Prettified (localized) value

    """
    if not request:
        request = get_current_request()

    # Under some circumstances, get_current_request will return None as
    # the current request. This makes it impossible to retrieve the
    # current locale or getting the translate method.
    if not request:
        locale_name = 'en_EN'
        _ = lambda t: t
    else:
        locale_name = get_locale_name(request)
        _ = request.translate

    if isinstance(value, datetime):
        return format_datetime(get_local_datetime(value),
                               locale_name=locale_name, format="short")
    elif isinstance(value, date):
        return format_date(value,
                           locale_name=locale_name, format="short")
    elif isinstance(value, timedelta):
        return from_timedelta(value)
    elif isinstance(value, list):
        values = []
        is_literal = False
        for v in value:
            r = prettify(request, v)
            if isinstance(r, literal):
                is_literal = True
            values.append(r)
        if is_literal:
            return literal(", ").join(values)
        else:
            return ", ".join(values)
    elif hasattr(value, "render"):
        return value.render()
    elif value is None:
        return ""
    return value

###########################################################################
#                               Times & Dates                             #
###########################################################################


def get_local_datetime(dt, timezone=None):
    """Will return a datetime converted into to given timezone. If the
    given datetime is naiv and does not support timezone information
    then UTC timezone is assumed. If timezone is None, then the local
    timezone of the server will be used.

    :dt: datetime
    :timezone: String timezone (eg. Europe/Berin)
    :returns: datetime

    """
    if not dt.tzinfo:
        dt = dt.replace(tzinfo=tz.gettz('UTC'))
    if not timezone:
        timezone = tz.tzlocal()
    return dt.astimezone(timezone)


def get_week(current_datetime):
    """Returns a tuple definig the start and end datetime for the week of
    the given date. Time from 00:00:00 -> 23:59:59"""
    last_day = 6
    current_day = current_datetime.weekday()
    _ws = current_datetime - timedelta(days=(current_day))
    _we = current_datetime + timedelta(days=(last_day - current_day))
    week_start = _ws.replace(hour=0, minute=0,
                             second=0, microsecond=0)
    week_end = _we.replace(hour=23, minute=59,
                           second=59, microsecond=0)
    return (week_start, week_end)


def format_timedelta(td):
    """Returns a formtted out put of a given timedelta in the form
    00:00:00"""
    if td < timedelta(0):
        return '-' + format_timedelta(-td)
    else:
        hours = td.total_seconds() // 3600
        minutes = (td.seconds % 3600) // 60
        seconds = td.seconds % 60
    return '%02d:%02d:%02d' % (hours, minutes, seconds)


def format_datetime(dt, locale_name=None, format="medium"):
    """Returns a prettyfied version of a datetime. If a locale_name is
    provided the datetime will be localized. Without a locale the
    datetime will be formatted into the form YYYY-MM-DD hh:ss"""
    if locale_name:
        locale = Locale(locale_name)
        return babel_format_datetime(dt, locale=locale, format=format)
    return dt.strftime("%Y-%m-%d %H:%M")


def format_date(dt, locale_name=None, format="medium"):
    """Returns a prettyfied version of a date. If a locale_name is
    provided the date will be localized. Without a locale the
    datetime will be formatted into the form YYYY-MM-DD hh:ss"""
    if locale_name:
        locale = Locale(locale_name)
        return babel_format_date(dt, locale=locale, format=format)
    return dt.strftime("%Y-%m-%d")



import os
from sqlalchemy.ext.declarative import declarative_base
from ringo.lib import helpers

# Define enabled ringo extenstion here. Extensions will be initialised
# on application startup.
extensions = [
]

# Global declarative base class.
Base = declarative_base()

# Dynamically import all model files and extensions here to have the
# model available in the sqlalchemy metadata. This is needed for the
# schema migration with alembic. Otherwise the migration will produce
# many table drops as the tables of the models are not present in the
# metadata
model_dir = os.path.dirname(os.path.realpath(__file__))
modul_files = [f for f in os.listdir(model_dir) if f.split(".")[-1] == "py"]
for model in modul_files:
    helpers.dynamic_import(__name__ + "." + model.split(".")[0])
for extension in extensions:
    helpers.dynamic_import(extension + ".model")



"""Modul for the messanging system in ringo"""
import logging
import datetime
import json
import csv
import codecs
import cStringIO
import sets
try:
        import cStringIO as StringIO
except ImportError:
        import StringIO
import xlsxwriter
import sqlalchemy as sa

from ringo.model.base import BaseItem
from ringo.model.user import UserSetting
from ringo.lib.helpers import serialize

log = logging.getLogger(__name__)


class ExtendedJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, BaseItem):
            return obj.id
            # Let the base class default method raise the TypeError
        elif isinstance(obj, datetime.datetime):
            return obj.strftime("%Y-%m-%d %H:%M:%S")
        elif isinstance(obj, datetime.date):
            return obj.strftime("%Y-%m-%d")
            # Let the base class default method raise the TypeError
        elif isinstance(obj, UserSetting):
            return obj.id
        else:
            return json.JSONEncoder.default(self, obj)


class RecursiveExporter(object):

    """Recurisve exporter for items of type clazz. For each item in the
    export the exporter will traverse down the relations of the items
    based on a given export configuration. If no export configuration is
    given the Exporter exports all fields of the items excluding
    relations and blobforms.

    Example export configuration::

    ["id", "foo", "bar", {"baz": ["f1", "f2", {"r1": [...]}, "baq": [...]}]

    In this configuration "id", "foo", "bar"... "f2" are considered as
    fields of the exported items. In contrast the keys of the nested
    dictionarys are taken as the name of the relations. The following
    list defines again the fields of the items in the relation.
    """

    def __init__(self, clazz, config):
        """@todo: to be defined1.

        :clazz: Clazz
        :config: Export configuration

        """
        if config:
            self._config = self._parse_config(json.loads(config))
        else:
            self._config = {"root": None}
        self._clazz = clazz
        self._data = {}

    def _parse_config(self, config, relation="root", result=None):
        """Will return a dictionary with the field configuration for each
        relation found in the export configuration. The field configuration
        is a list of fieldnames.

        :config: Export configuration
        :relation: Name of the "current" relation. The name "root" is a
        placeholder for the elements on the first level of the export.
        :result: temporary relation configuration. Is neeed for recursion in
        this function
        :returns: Dict with realtion configuration

        """
        if result is None:
            result = {}
        result[relation] = []
        for field in config:
            if isinstance(field, dict):
                for rel in field:
                    result[rel] = self._parse_config(field[rel],
                                                     rel,
                                                     result)[rel]
                    result[relation].append(rel)
            else:
                result[relation].append(field)
        return result

    def _iter_export(self, export, relation_config, relation):
        raise NotImplementedError()

    def _remove_duplicates(self):
        raise NotImplementedError()

    def perform(self, items):
        """@todo: Docstring for perform.
        function

        :arg1: @todo
        :returns: @todo

        """
        fields_to_export = self._config["root"]
        exporter = Exporter(self._clazz, fields_to_export, serialized=False)
        export = exporter.perform(items)
        self._iter_export(export, self._config, "root")
        self._remove_duplicates()
        return self._data


class RecursiveRelationExporter(RecursiveExporter):

    """Recursive Expoert for items of type clazz. After the export has
    finished the exported data will are returned as a dictionary. Each
    key in the dictionary will hold the exported values of the
    configured relations in the export configuration. The items of type
    clazz acn be found in the dict under the key 'root'"""

    def get_relation_config(self):
        return self._config

    def _iter_export(self, export, relation_config, relation):
        if self._data.get(relation) is None:
            self._data[relation] = []
        for item in export:
            temp = {}
            for field in item:
                if field in relation_config:
                    fields_to_export = relation_config[field]
                    if isinstance(item[field], list) and len(item[field]) > 0:
                        clazz = item[field][0].__class__
                        exporter = Exporter(clazz,
                                            fields_to_export,
                                            serialized=False)
                        self._iter_export(exporter.perform(item[field]),
                                          relation_config,
                                          field)
                    elif item[field] is not None:
                        clazz = item[field]
                        exporter = Exporter(clazz,
                                            fields_to_export,
                                            serialized=False)
                        self._iter_export(exporter.perform([item[field]]),
                                          relation_config,
                                          field)
                else:
                    temp[field] = item[field]
            self._data[relation].append(temp)

    def _remove_duplicates(self):
        for relation in self._data:
            visited = []
            unique = []
            for item in self._data[relation]:
                hashv = hash(unicode(item))
                if hashv not in visited:
                    visited.append(hashv)
                    unique.append(item)
                else:
                    continue
            self._data[relation] = unique


class UnicodeCSVWriter:
    """
    A CSV writer which will write rows to CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(self, f, fields, dialect=csv.excel, encoding="utf-8", **kwds):
        # Redirect output to a queue
        self.queue = cStringIO.StringIO()
        self.writer = csv.DictWriter(self.queue, fieldnames=fields,
                                     dialect=dialect, **kwds)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)()

    def writeheader(self):
        self.writer.writeheader()

    def writerow(self, row):
        tmp_dict = {}
        for k, v in row.iteritems():
            try:
                tmp_dict[k] = v.encode("utf-8")
            except AttributeError:
                tmp_dict[k] = v
        self.writer.writerow(tmp_dict)
        # Fetch UTF-8 output from the queue ...
        data = self.queue.getvalue()
        data = data.decode("utf-8")
        # ... and reencode it into the target encoding
        data = self.encoder.encode(data)
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)


class Exporter(object):

    """Docstring for Exporter. """

    def __init__(self, clazz, fields=None, serialized=True, relations=False):
        """Base exporter to export items of the given class. The
        exporter will return a list of dictionarys with key values pairs
        of the values for each items which should be exported. The
        fields to be exported can be configured. You can configure if
        the values should be serialized too.

        On default no relations will be exported. This can be changed by
        either setting the relations flag to true or defining relations
        explicit in the fields attribute.

        Exported relations will the id of the linked items.

        :clazz: Clazz of the items which will be exported
        :fields: List of fields and relations which should be exported.
        If no fields are provided all fields excluding the relations
        will be exported. The order of the configured fields will
        determine the order of the fields in the export (If supported
        e.g CSV).
        :serialized: Flag to indicate that the exported values should be
        serialized e.g dates will be converted into ISO8601 format. If
        realtions are included in the export the serialzed form will be
        the string representation of the item.
        Defaults to True.

        """
        self._clazz = clazz
        self._fields = fields
        self._serialized = serialized
        self._relations = relations

    def serialize(self, data):
        """Will convert the given python data dictionary into a string
        containing JSON data

        :data: Dictionary containing Python data.
        :returns: String representing the data

        """
        return data

    def flatten(self, data):
        """Will flatten nested data structures as found in Blobforms"""
        values = {}
        for key in data:
            # Handle data container of blobforms
            if key == "data":
                jdata = json.loads(data[key])
                for jkey in jdata:
                    values[jkey] = jdata[jkey]
            else:
                values[key] = data[key]
        return values

    def perform(self, items):
        """Returns the serialized item as string

        :items: @todo
        :returns: @todo

        """
        data = []
        for item in items:
            # Ensure that every item has a UUID. Set missing UUID here
            # if the item has no uuid set yet.
            if not item.uuid:
                item.reset_uuid()
            if self._fields is None:
                # Default export. Export all fields excluding relations
                values = item.get_values(serialized=self._serialized,
                                         include_relations=self._relations)
            else:
                values = {}
                for field in self._fields:
                    value = item.get_value(field)
                    if self._serialized:
                        value = serialize(value)
                    values[field] = value
            data.append(self.flatten(values))
        return self.serialize(data)


class XLSXExporter(Exporter):
    """Docstring for XLSXExporter. """

    def serialize(self, data):
        output = StringIO.StringIO()
        book = xlsxwriter.Workbook(output)
        if len(data) > 0:
            keys = sorted(data[0].keys())
            sheet = book.add_worksheet(self._clazz.__tablename__)
            row = 0
            col = 0
            # write header
            for key in keys:
                sheet.write(row, col, key)
                col += 1
            # write data
            col = 0
            row = 1
            for item in data:
                for key in keys:
                    sheet.write(row, col, item[key])
                    col += 1
                row += 1
                col = 0
        book.close()
        output.seek(0)
        return output.read()


class JSONExporter(Exporter):
    """Docstring for JSONExporter. """

    def serialize(self, data):
        return json.dumps(data, cls=ExtendedJSONEncoder)


class CSVExporter(Exporter):
    """Docstring for CSVExporter. """

    def _collect_keys(self, data):
        """The function will collect all keys (fields) within the given
        items. This is needed in case of blobform items as those items
        has a generic data field which may contain variable number of
        fields depending if the item has a value for the fields or
        not."""
        keys = sets.Set()
        for item in data:
            keys = keys.union(item.keys())
        return keys

    def serialize(self, data):
        outfile = cStringIO.StringIO()
        writer = UnicodeCSVWriter(outfile, sorted(self._collect_keys(data)))
        writer.writeheader()
        writer.writerows(data)
        outfile.seek(0)
        return outfile.read()


class Importer(object):
    """Docstring for Importer."""

    def __init__(self, clazz, db=None):
        """@todo: to be defined1.

        :clazz: The clazz for which we will import data

        """
        self._clazz = clazz
        self._db = db
        self._clazz_type = self._get_types(clazz)

    def _get_types(self, clazz):
        type_mapping = {}
        mapper = sa.orm.class_mapper(clazz)
        for prop in mapper.iterate_properties:
            if isinstance(prop, sa.orm.RelationshipProperty):
                type_mapping[prop.key] = str(prop.direction.name)
            else:
                type_mapping[prop.key] = str(prop.columns[0].type)
        return type_mapping

    def _deserialize_values(self, obj):
        """This function can be called after the basic deserialisation
        has finished. It is used to convert integer, date and datetime
        objects which are either not supported by the defaults decoders
        or not decoded correct (NULL values)

        :obj: Deserialized dictionary from basic deserialisation
        :returns: Deserialized dictionary with additional integer, date
        and datetime deserialisation
        """
        for field in obj:
            if (not field in self._clazz_type
               or not self._clazz_type[field] in ['DATE',
                                                  'DATETIME',
                                                  'INTEGER']
               or obj[field] is None):
                continue
            elif obj[field] == "":
                obj[field] = None
                continue
            elif self._clazz_type[field] == "INTEGER":
                obj[field] = int(obj[field])
            elif self._clazz_type[field] == "DATE":
                obj[field] = datetime.datetime.strptime(
                    obj[field], "%Y-%m-%d").date()
            elif self._clazz_type[field] == "DATETIME":
                obj[field] = datetime.datetime.strptime(
                    obj[field], "%Y-%m-%d %H:%M:%S")
        return obj

    def _deserialize_relations(self, obj):
        """Will deserialize items in a MANYTOMANY relation. Other
        relations do not need to be handled as they should have a
        foreign key to the related item which is part of the items field
        anyway. It will replace the id values of the related items with
        the loaded items. This only works if there is a db connection
        available.

        :obj: Deserialized dictionary from basic deserialisation
        :returns: Deserialized dictionary with additional MANYTOMANY
        relations.
        """
        for field in obj.keys():
            try:
                ftype = self._clazz_type[field]
            except KeyError:
                log.warning("Can not find field %s in %s" % (field, self._clazz_type))
                continue
            # Handle all types of relations...
            if ftype in ["MANYTOMANY", "MANYTOONE",
                         "ONETOONE", "ONETOMANY"]:
                # Remove the items from the list if there is no db
                # connection or of there are not MANYTOMANY.
                if not self._db or (ftype != "MANYTOMANY"):
                    del obj[field]
                    continue
                clazz = getattr(self._clazz, field).mapper.class_
                tmp = []
                for item_id in obj[field]:
                    q = self._db.query(clazz).filter(clazz.id == item_id)
                    try:
                        tmp.append(q.one())
                    except:
                        log.warning(("Can not load '%s' id: %s "
                                     "Relation '%s' of '%s' not set"
                                     % (clazz, item_id, field, self._clazz)))
                obj[field] = tmp
        return obj

    def deserialize(self, data):
        """Will convert the string data into a dictionary like data.

        :data: Importdata as string (JSON, XML...)
        :returns: Dictionary like data

        """
        return {}

    def perform(self, data, user=None, translate=lambda x: x, use_uuid=True):
        """Will return a list of imported items. The list will contain a
        tupel of the item and a string which gives information on the
        operaten (update, create). For create operations the new item
        will be created with the given user.

        :data: Importdata as string (JSON, XML...)
        :user: User object. Used when creating objects.
        :translate: Translation method.
        :use_uuid: Loading items by their uuid.
        :returns: List of imported items

        """
        imported_items = []
        import_data = self.deserialize(data)
        factory = self._clazz.get_item_factory()
        _ = translate
        for values in import_data:
            if use_uuid:
                id = values.get('uuid')
                if "id" in values:
                    del values["id"]
            else:
                id = values.get('id')
            try:
                # uuid might be empty for new items, which will raise an
                # error on loading.
                item = factory.load(id or "thisiddoesnotexist",
                                    uuid=use_uuid)
                item.set_values(values)
                operation = _("UPDATE")
            except:
                item = factory.create(user=user, values=values)
                operation = _("CREATE")
            imported_items.append((item, operation))
        return imported_items


class JSONImporter(Importer):
    """Docstring for JSONImporter."""

    def _deserialize_hook(self, obj):
        obj = self._deserialize_values(obj)
        return self._deserialize_relations(obj)

    def deserialize(self, data):
        """Will convert the JSON data back into a dictionary with python values

        :data: String JSON data
        :returns: List of dictionary with python values
        """
        conv = json.loads(data, object_hook=self._deserialize_hook)
        if isinstance(conv, dict):
            return [conv]
        return conv


class CSVImporter(Importer):
    """Docstring for CSVImporter."""

    def _deserialize_hook(self, obj):
        conv = {}
        for k, v in obj.iteritems():
            conv[k] = unicode(v, "utf-8")
        conv = self._deserialize_values(conv)
        return conv

    def deserialize(self, data):
        """Will convert the CSV data back into a dictionary with python values

        :data: String CSV data
        :returns: List of dictionary with python values
        """
        result = []
        infile = cStringIO.StringIO(data)
        reader = csv.DictReader(infile)
        for conv in reader:
            conv = self._deserialize_hook(conv)
            result.append(conv)
        return result



import uuid
import logging

import pyramid.httpexceptions as exc
from pyramid.security import remember, forget
from pyramid.view import view_config
from pyramid.httpexceptions import HTTPFound

from formbar.config import Config, load
from formbar.form import Form, Validator

from ringo.lib.sql import DBSession
from ringo.model.user import USER_GROUP_ID, USER_ROLE_ID
from ringo.lib.helpers import import_model
from ringo.views.request import handle_history
from ringo.views.users import (
    password_minlength_validator,
    password_nonletter_validator
)
from ringo.lib.helpers.appinfo import get_app_title
from ringo.lib.helpers.misc import dynamic_import
from ringo.lib.form import get_path_to_form_config
from ringo.lib.security import login as user_login, request_password_reset, \
    password_reset, activate_user, encrypt_password, AuthentificationException
from ringo.lib.message import Mailer, Mail

User = import_model('ringo.model.user.User')
Usergroup = import_model('ringo.model.user.Usergroup')
Role = import_model('ringo.model.user.Role')

log = logging.getLogger(__name__)


def is_login_unique(field, data):
    """Validator function as helper for formbar validators"""
    users = DBSession.query(User).filter_by(login=data[field]).all()
    return len(users) == 0


def is_registration_enabled(settings):
    return (bool(settings.get('mail.host')) and
            bool(settings.get('mail.default_sender')) and
            settings.get('auth.register_user') == "true")


def is_pwreminder_enabled(settings):
    return (bool(settings.get('mail.host')) and
            bool(settings.get('mail.default_sender')) and
            settings.get('auth.password_reminder') == "true")


def is_authcallback_enabled(settings):
    return bool(settings.get('auth.callback'))


@view_config(route_name='login', renderer='/auth/login.mako')
def login(request):
    handle_history(request)
    _ = request.translate
    settings = request.registry.settings
    config = Config(load(get_path_to_form_config('auth.xml')))
    form_config = config.get_form('loginform')
    form = Form(form_config, csrf_token=request.session.get_csrf_token(),
                translate=_)
    if request.POST:
        form.validate(request.params)
        username = form.data.get('login')
        password = form.data.get('pass')
        user = user_login(username, password)
        if user is None:
            msg = _("Login failed!")
            request.session.flash(msg, 'error')
        elif not user.activated:
            msg = _("Login failed!")
            request.session.flash(msg, 'error')
            target_url = request.route_path('accountdisabled')
            return HTTPFound(location=target_url)
        else:

            # Handle authentication callback.
            if is_authcallback_enabled(settings):
                authenticated = False
                try:
                    callback = dynamic_import(settings.get("auth.callback"))
                    callback(request, user)
                    authenticated = True
                except AuthentificationException as e:
                    msg = e.message
                    request.session.flash(msg, 'critical')
            else:
                authenticated = True

            if authenticated:
                msg = _("Login was successfull")
                request.session.flash(msg, 'success')
                headers = remember(request, user.id)
                target_url = request.route_path('home')
                return HTTPFound(location=target_url, headers=headers)

    return {'form': form.render(),
            'registration_enabled': is_registration_enabled(settings),
            'pwreminder_enabled': is_pwreminder_enabled(settings)}


@view_config(route_name='logout', renderer='/auth/logout.mako')
def logout(request):
    handle_history(request)
    _ = request.translate
    target_url = request.route_path('home')
    if request.params.get('autologout'):
        target_url = request.route_path('autologout')
        return HTTPFound(location=target_url)
    elif request.user:
        log.info("Logout successfull '%s'" % (request.user.login))
        msg = _("Logout was successfull")
        headers = forget(request)
        request.session.flash(msg, 'success')
        return HTTPFound(location=target_url, headers=headers)
    return HTTPFound(location=target_url)


@view_config(route_name='autologout', renderer='/auth/autologout.mako')
def autologout(request):

    # For the first call the user is still authenticated. So
    # delete the auth cookie and trigger a redirect calling the same
    # page.
    if request.user:
        headers = forget(request)
        target_url = request.route_path('autologout')
        log.info("Autologout successfull '%s'" % (request.user.login))
        return HTTPFound(location=target_url, headers=headers)

    # User is not authenticated here anymore. So simply render the
    # logout page.
    _ = request.translate
    return {"_": _}


@view_config(route_name='accountdisabled', renderer='/auth/disabled.mako')
def accountdisabled(request):
    _ = request.translate
    return {"_": _}


@view_config(route_name='register_user',
             renderer='/auth/register_user.mako')
def register_user(request):
    settings = request.registry.settings
    if not is_registration_enabled(settings):
        raise exc.exception_response(503)
    handle_history(request)
    _ = request.translate
    config = Config(load(get_path_to_form_config('auth.xml')))
    form_config = config.get_form('register_user')
    form = Form(form_config, csrf_token=request.session.get_csrf_token(),
                translate=_)
    # Do extra validation which is not handled by formbar.
    # Is the login unique?
    login_unique_validator = Validator('login',
                                       _('There is already a user with this '
                                         'name'),
                                       is_login_unique)
    pw_len_validator = Validator('pass',
                                 _('Password must be at least 12 characters '
                                   'long.'),
                                 password_minlength_validator)
    pw_nonchar_validator = Validator('pass',
                                     _('Password must contain at least 2 '
                                       'non-letters.'),
                                     password_nonletter_validator)
    form.add_validator(login_unique_validator)
    form.add_validator(pw_len_validator)
    form.add_validator(pw_nonchar_validator)
    registration_complete = False
    if request.POST:
        if form.validate(request.params):
            # 1. Create user. Do not activate him. Default role is user.
            ufac = User.get_item_factory()
            user = ufac.create(None, form.data)
            # Set login from formdata
            user.login = form.data['login']
            # Encrypt password and save
            user.password = encrypt_password(form.data['pass'])
            # Deactivate the user. To activate the user needs to confirm
            # with the activation link
            user.activated = False
            atoken = str(uuid.uuid4())
            user.activation_token = atoken
            # Set profile data
            user.profile[0].email = form.data['_email']

            # 2. Set user group
            gfac = Usergroup.get_item_factory()
            default_grps = settings.get("auth.register_user_default_groups",
                                        str(USER_GROUP_ID))
            for gid in [int(id) for id in default_grps.split(",")]:
                group = gfac.load(gid)
                user.groups.append(group)

            # 3. Set user role
            rfac = Role.get_item_factory()
            default_roles = settings.get("auth.register_user_default_roles",
                                         str(USER_ROLE_ID))
            for rid in [int(id) for id in default_roles.split(",")]:
                role = rfac.load(rid)
                user.roles.append(role)
            # Set default user group.
            request.db.add(user)

            # 4. Send confirmation email. The user will be activated
            #    after the user clicks on the confirmation link
            mailer = Mailer(request)
            recipient = user.profile[0].email
            subject = _('Confirm user registration')
            values = {'url': request.route_url('confirm_user', token=atoken),
                      'app_name': get_app_title(),
                      'email': settings['mail.default_sender'],
                      '_': _}
            mail = Mail([recipient],
                        subject,
                        template="register_user",
                        values=values)
            mailer.send(mail)

            msg = _("User has been created and a confirmation mail was sent"
                    " to the users email adress. Please check your email.")
            request.session.flash(msg, 'success')
            registration_complete = True
    return {'form': form.render(), 'complete': registration_complete}


@view_config(route_name='confirm_user',
             renderer='/auth/confirm_user.mako')
def confirm_user(request):
    settings = request.registry.settings
    if not is_registration_enabled(settings):
        raise exc.exception_response(503)
    handle_history(request)
    _ = request.translate
    success = False
    token = request.matchdict.get('token')
    user = activate_user(token, request.db)
    if user:
        success = True
        msg = _("The user has beed successfull confirmed.")
    else:
        msg = _("The user was not confirmed. Maybe the confirmation"
                " token was not valid?")
    return {'msg': msg, 'success': success}


@view_config(route_name='forgot_password',
             renderer='/auth/forgot_password.mako')
def forgot_password(request):
    settings = request.registry.settings
    if not is_pwreminder_enabled(settings):
        raise exc.exception_response(503)
    handle_history(request)
    _ = request.translate
    config = Config(load(get_path_to_form_config('auth.xml')))
    form_config = config.get_form('forgot_password')
    form = Form(form_config, csrf_token=request.session.get_csrf_token(),
                translate=_)
    complete = False
    if request.POST:
        if form.validate(request.params):
            username = form.data.get('login')
            user = request_password_reset(username, request.db)
            if user:
                mailer = Mailer(request)
                recipient = user.profile[0].email
                token = user.reset_tokens[-1]
                subject = _('Password reset request')
                values = {'url': request.route_url('reset_password',
                                                   token=token),
                          'app_name': get_app_title(),
                          'email': settings['mail.default_sender'],
                          '_': _}
                mail = Mail([recipient],
                            subject,
                            template="password_reset_request",
                            values=values)
                mailer.send(mail)
            msg = _("Password reset token has been sent to the users "
                    "email address. Please check your email.")
            request.session.flash(msg, 'success')
            complete = True
    return {'form': form.render(), 'complete': complete}


@view_config(route_name='reset_password',
             renderer='/auth/reset_password.mako')
def reset_password(request):
    settings = request.registry.settings
    if not is_pwreminder_enabled(settings):
        raise exc.exception_response(503)
    handle_history(request)
    _ = request.translate
    success = False
    token = request.matchdict.get('token')
    user, password = password_reset(token, request.db)
    if password:
        mailer = Mailer(request)
        recipient = user.profile[0].email
        subject = _('Password has been reseted')
        values = {'password': password,
                  'app_name': get_app_title(),
                  'email': settings['mail.default_sender'],
                  '_': _}
        mail = Mail([recipient],
                    subject,
                    template="password_reminder",
                    values=values)
        mailer.send(mail)
        msg = _("Password was resetted and sent to the users email address."
                " Please check your email.")
        success = True
    else:
        msg = _("Password was not resetted. Maybe the request"
                " token was not valid?")
    return {'msg': msg, 'success': success}



import logging
from pyramid.view import view_config

from ringo.lib.helpers import get_action_routename
from ringo.lib.form import get_path_to_form_config
from ringo.views.base import create
from ringo.model.form import Form

log = logging.getLogger(__name__)


def load_config(path):
    data = ""
    with open(path) as f:
        data = f.read()
    return data


@view_config(route_name=get_action_routename(Form, 'create'),
             renderer='/default/create.mako',
             permission='create')
def create_(request):
    clazz = request.context.__model__
    factory = clazz.get_item_factory()
    form = factory.create(request.user, values={})
    definition = load_config(get_path_to_form_config("blobform_template.xml",
                                                     "ringo"))
    form.definition = definition
    request.context.item = form
    return create(request)



def secure_headers_factory(handler, registry):
    def secure_headers_tween(request):
        response = request.response
        # Just modify this dict to inject new headers
        HEADERS = {
            'X-XSS-Protection': '1; mode=block',
            'X-Content-Type-Options': 'nosniff'
        }
        for header in HEADERS:
            if not response.headers.get(header, None):
                response.headers[header] = HEADERS[header]
        return handler(request)
    return secure_headers_tween



import urlparse


class History:

    def __init__(self, history):
        self.history = history

    def push(self, url):
        """Adds an url to the history if the url is not already the most
        recent entry. The scheme and network location (host, port,
        username, password), if present, are removed from the URL before
        storing it. If there are more than 5 entries in the list the
        oldes entry will be removed.
        """

        # normalize the URL by removing scheme and netloc. This avoids
        # problems with the URLs when running ringo behind reverse
        # proxies.
        split = urlparse.urlsplit(url)
        normalized_url = urlparse.urlunsplit(("", "") + split[2:])

        if not self.history or normalized_url != self.history[-1]:
            self.history.append(normalized_url)
        if len(self.history) > 5:
            del self.history[0]

    def pop(self, num=1):
        """Returns a url form the history and deletes the item and all
        decendants from the history. On default it will return the last
        recent entry in the history. Optionally you can provide a number
        to the pop method to get e.g the 2 most recent entry."""
        url = None
        for x in range(num):
            if len(self.history) > 0:
                url = self.history.pop()
        return url

    def last(self):
        """Returns the last element from the history stack without
        removing it"""
        if len(self.history) > 0:
            return self.history[-1]
        return None



import logging
import os
import pkg_resources
from mako.lookup import TemplateLookup
import ringo.lib.helpers
from ringo.lib.helpers import (
    get_app_name,
    get_saved_searches,
    get_item_actions,
    literal
)
from ringo.lib.table import get_table_config
import ringo.lib.security as security

base_dir = pkg_resources.get_distribution("ringo").location
template_dir = os.path.join(base_dir, 'ringo', 'templates')
template_lookup = TemplateLookup(directories=[template_dir],
                                 default_filters=['h'])

log = logging.getLogger(__name__)

###########################################################################
#                         Renderers for overviews                         #
###########################################################################


class ListRenderer(object):
    """Docstring for ListRenderer """

    def __init__(self, listing):
        """@todo: to be defined """
        self.listing = listing
        self.config = get_table_config(self.listing.clazz)
        self.template = template_lookup.get_template("internal/list.mako")

    def render(self, request):
        """Initialize renderer"""
        # TODO: Enabled sorting of lists. Mind that these lists might be
        # presorted if the user clicked on the header. In this case some
        # get params with sort configurations are in the session. This
        # logic is currently in base/view. (ti) <2014-01-23 23:15>
        # sort_field = self.config.get_default_sort_column()
        # sort_order = self.config.get_default_sort_order()
        # self.listing.sort(sort_field, sort_order)

        if len(self.listing.search_filter) > 0:
            search = self.listing.search_filter[-1][0]
            search_field = self.listing.search_filter[-1][1]
            regexpr = self.listing.search_filter[-1][2]
        else:
            search = ""
            search_field = ""
            regexpr = False
        ssearch = get_saved_searches(request,
                                     self.listing.clazz.__tablename__)

        bundled_actions = []
        for action in get_item_actions(request, self.listing.clazz):
            if action.bundle and security.has_permission(action.name.lower(),
                                                         request.context,
                                                         request):
                bundled_actions.append(action)

        values = {'items': self.listing.items,
                  'clazz': self.listing.clazz,
                  'listing': self.listing,
                  'request': request,
                  '_': request.translate,
                  'h': ringo.lib.helpers,
                  's': security,
                  'bundled_actions': bundled_actions,
                  'search': search,
                  'regexpr': regexpr,
                  'search_field': search_field,
                  'saved_searches': ssearch,
                  'tableconfig': self.config}
        return literal(self.template.render(**values))


class DTListRenderer(object):
    """Docstring for a ListRenderer using the DataTables Jquery Plugin"""
    def __init__(self, listing, tablename=None):
        self.listing = listing
        self.config = get_table_config(self.listing.clazz, tablename)
        self.template = template_lookup.get_template("internal/dtlist.mako")

        # Template for JS datatable configuration
        self.js_template = template_lookup.get_template("internal/dtlist.js.mako")


    def _get_table_id(self):
        table_id = ["dt",
                    self.config.clazz.__tablename__,
                    self.config.name]
        table_id = "_".join(table_id)
        return table_id


    def _render_js_config(self, request, table_id, bundled_actions):
        values = {'tableconfig': self.config,
                  'table_id': table_id,
                  'request': request,
                  'bundled_actions': bundled_actions,
                  '_': request.translate}
        return self.js_template.render(**values)

    def render(self, request):
        """Initialize renderer"""

        # Get the bundles_actions.
        # If bundle_actions is True, display checkboxes and more in the table.
        bundled_actions = []
        for action in get_item_actions(request, self.listing.clazz):
            if action.bundle and security.has_permission(action.name.lower(),
                                                         request.context,
                                                         request):
                bundled_actions.append(action)
        table_id = self._get_table_id()
        table_config = self._render_js_config(request,
                                              table_id,
                                              bundled_actions)
        values = {'items': self.listing.items,
                  'clazz': self.listing.clazz,
                  'listing': self.listing,
                  'request': request,
                  '_': request.translate,
                  's': security,
                  'h': ringo.lib.helpers,
                  'bundled_actions': bundled_actions,
                  'tableconfig': self.config,
                  'tableid': table_id,
                  'dtconfig': table_config}
        return literal(self.template.render(**values))


class StaticListRenderer(ListRenderer):

    """Docstring for StaticListRendenderer."""

    def __init__(self, listing, tablename=None):
        self.listing = listing
        self.config = get_table_config(self.listing.clazz, tablename)
        self.template = template_lookup.get_template("internal/staticlist.mako")

    def _get_table_id(self):
        table_id = ["static",
                    self.config.clazz.__tablename__,
                    self.config.name]
        table_id = "_".join(table_id)
        return table_id

    def render(self, request):
        """Initialize renderer"""

        table_id = self._get_table_id()
        values = {'items': self.listing.items,
                  'clazz': self.listing.clazz,
                  'listing': self.listing,
                  'request': request,
                  '_': request.translate,
                  's': security,
                  'h': ringo.lib.helpers,
                  'tableconfig': self.config,
                  'tableid': table_id}
        return literal(self.template.render(**values))



"""The api modul include view functions which are some sort of helper
functions usually called by the client"""
import logging
from pyramid.response import Response
from pyramid.view import view_config

from formbar.form import Form
from formbar.config import Config, parse
from formbar.rules import Rule

from ringo.views.response import JSONResponse
from ringo.views.helpers import set_current_form_page as save_form_page

log = logging.getLogger(__name__)


@view_config(route_name='keepalive',
             renderer='json',
             request_method="GET")
def keepalive(request):
    """Endpoint for simple requests to trigger refreshing the logouttimer"""
    return JSONResponse(True, "I'm alive!", {})


@view_config(route_name='rules-evaluate',
             renderer='json',
             request_method="GET")
def evaluate(request):
    """Method which will evaluate a formed rule given in the GET
    request. It will return a JSONResponse with the result of the
    evaluation."""
    try:
        ruleexpr = request.GET.get('rule').strip()
        rule = Rule(ruleexpr)
        result = rule.evaluate({})
        return JSONResponse(True, result, {"msg": rule.msg})
    except:
        msg = "Can not evaluate rule '%s'" % ruleexpr
        log.error(msg)
        return JSONResponse(False, False, {"msg": msg})


@view_config(route_name='form-render',
             renderer='json',
             request_method="POST")
def render(request):
    """Will return a JSONResponse with a rendererd form. The form
    defintion and the formid is provided in the POST request."""
    _ = request.translate
    form_config = request.POST.get("definition")
    config_name = request.POST.get("formid")
    out = []
    try:
        config = Config(parse(form_config))
        form_config = config.get_form(config_name)
        form = Form(form_config, None, request.db,
                    csrf_token=request.session.get_csrf_token())
        out.append(form.render(buttons=False, outline=False))
    except Exception as ex:
        out.append(_('ERROR: %s' % str(ex)))
    data = {"form": "".join(out)}
    return JSONResponse(True, data, {"msg": "Ole!"})


@view_config(route_name='set_current_form_page')
def set_current_form_page(request):
    """Will save the currently selected page of a form in the session.
    The request will have some attributes in the GET request which will
    config which page, of which item is currently shown. This function
    is used as a callback function within formbar.

    :request: Current request
    :returns: Response
    """
    page = request.GET.get('page')
    tablename = request.GET.get('item')
    itemid = request.GET.get('itemid')
    if page and tablename and itemid:
        save_form_page(tablename, itemid, page, request)
    return Response(body='OK', content_type='text/plain')



"""
Extensions are external plugable modules. They are used to add generic
functionallity to the application in a dynamic way.

In contrast to modules, extensions do not implement application specific
stuff, but are generix with functionallity which is usefull for all kind
of applications. An example might be an appointment extensions which
allows to extend the application with an appointment feature.

For more information on how to create and register extensions please
refer to the :ref:`extensions` part of the documentation.

Extension may extend the existing database and adds tables to save data
(e.g appointments and relations of appointments to users.) or they are
lightweight and just provides some functionallity like anonymisation.

Extension can be Registered to an application by adding the extentions
in the `extensions` list located in the `modul/__init__.py` file. As
soon as the application is started the extension will be registered.
To unregister a extension simply remove the extension from the
`extentions` list.

.. note::
    For extensions which modifies the datamodel of the application a
    seperate model migration is needed on registration and
    unregistration of the extension. (See alembic and ringo-admin
    documentation).


"""
import logging
import sqlalchemy as sa
import transaction
from ringo.lib.sql.db import DBSession
from ringo.model.modul import ModulItem, ACTIONS

log = logging.getLogger(name=__name__)


def _configure_actions(modul, config, actions, dbsession):
    """Will configure the configured actions of the modul in the
    database. The function will create and delete actions depending of
    the current modul configuration.

    :modul: Modul item of the extension
    :config: Extension configuration as a dictionary
    :actions: List of ActionItems
    :dbsession: Database session
    """
    # Add custom actions to the global ACTION dictionary
    if actions is None:
        actions = []
    for action in actions:
        ACTIONS[action.name.lower()] = action

    new_actions = config.get("actions") or []
    old_actions = [action.name.lower() for action in modul.actions]
    delete_actions = (set(old_actions) - set(new_actions))
    add_actions = (set(new_actions) - set(old_actions))
    for name in delete_actions:
        log.debug("Deleting action %s from modul %s" % (name, modul))
        action = modul.get_action(name)
        if action:
            modul.actions.remove(action)
    for name in add_actions:
        log.debug("Adding action %s to modul %s" % (name, modul))
        action = ACTIONS.get(name)
        if action:
            modul.actions.append(action)
        else:
            log.error("Action %s is not known!" % (name))
    return modul


def _add_modul(config, actions, dbsession):
    """Method to add the modul entries for the extension in the
    database. This includes a entry in the modules table and the
    configured actions.

    :config: Extension configuration as a dictionary
    :actions: List of ActionItems
    :dbsession: Database session
    """
    # Set defaults
    name = config.get("name")
    clazzpath = config.get("clazzpath")
    str_repr = config.get("str_repr") or "%s|id"
    display = config.get("display") or "hidden"
    label = config.get("label") or name
    label_plural = config.get("label_plural") or label

    # Add modul
    modul = ModulItem(name=name)
    modul.clazzpath = clazzpath
    modul.label = label
    modul.str_repr = str_repr
    modul.label_plural = label_plural
    modul.display = display
    # Configure actions
    modul = _configure_actions(modul, config, actions, dbsession)
    dbsession.add(modul)
    dbsession.flush()
    transaction.commit()
    log.debug("Adding new modul '%s'" % name)
    return modul


def _load_modul_by_name(session, modulname):
    try:
        # FIXME:
        # Compatibilty mode. Older versions of Ringo added a 's' to the
        # extensions modul name as Ringo usally uses the plural form.
        # Newer versions use the configured extension name. So there
        # might be a mixture of old and new modul names in the database.
        # This code will handle this. (ti) <2016-01-04 13:50>
        return session.query(ModulItem)\
               .filter(sa.or_(ModulItem.name == modulname,
                              ModulItem.name == modulname + 's'))\
                .one()
    except sa.orm.exc.NoResultFound:
        return None


def register_modul(config, modul_config, actions=None):
    return check_register_modul(config, modul_config, actions)


def check_register_modul(config, modul_config, actions=None):
    """This method is used by extension to register their modul with a
    given config in the application. This function will check if the
    modul is already registered. Registered in this context means that
    the modul has an entry in the modules table

    If it is not registered it will show a warning message. So the
    function really does not register the modul but check if it is
    registered.

    :config: Application config
    :modul_config: Extension configuration as a dictionary
    :actions: List of ActionItems
    :returns: ModulItem or None
    """

    modulname = modul_config.get('name')
    modul = _load_modul_by_name(DBSession, modulname)
    if modul:
        log.info("Extension '%s' OK" % modulname)
    else:
        log.warning("Extension '%s' is missing a entry in DB. "
                    "Missing migration?" % modulname)
    return modul


def check_unregister_modul(modul, extensions):
    """Will check if there is a entry in the moduls table for an
    extension which is not currently added in the application and can
    therefor be deleted. If there is such an entry a warning is logged.
    """
    app_name = modul.clazzpath.split(".")[0]
    if app_name.find("ringo_") > -1 and app_name not in extensions:
        log.warning("Spare entry in DB for Extension '%s'. "
                    "Missing migration?" % modul.name)



from pyramid.events import BeforeRender

import ringo.lib.security as security
import ringo.lib.helpers as helpers
from ringo.lib.form import (
    formbar_js_filenames,
    formbar_css_filenames
)
from ringo.lib.renderer.form import add_renderers
from ringo.lib.renderer.dialogs import (
    DialogRenderer,
    ConfirmDialogRenderer,
    ErrorDialogRenderer,
    WarningDialogRenderer,
    InfoDialogRenderer,
    ExportDialogRenderer,
    ImportDialogRenderer
)
from ringo.lib.renderer.lists import (
    ListRenderer,
    DTListRenderer
)


def setup_render_globals(config):
    config.add_subscriber(add_renderer_globals, BeforeRender)


def add_renderer_globals(event):
    request = event['request']
    default_locale = request.registry.settings.get("pyramid.default_locale_name", "en")
    event['h'] = helpers
    event['s'] = security
    event['_'] = request.translate
    event['N_'] = request.translate
    event['formbar_css_filenames'] = formbar_css_filenames
    event['formbar_js_filenames'] = formbar_js_filenames
    event['localizer'] = request.localizer
    event['client_language'] = getattr(request, "_LOCALE_", default_locale)
    event['application_path'] = helpers.get_app_url(request)



"""Functiont to work with forms."""
import os
import inspect
from formbar.form import Form
from formbar.config import Config, load, parse
from formbar.helpers import get_css_files, get_js_files
from ringo.lib.cache import CACHE_FORM_CONFIG
from ringo.lib.helpers import (
    get_path_to,
    get_app_inheritance_path
)

formbar_css_filenames = []
formbar_js_filenames = []


def get_eval_url():
    """Returns the REST API endpoint for form evaluation"""
    return '/rest/rule/evaluate'


def get_ownership_form(item, db, csrf_token, eval_url,
                       readonly=None, url_prefix=None, locale=None,
                       translate=None):
    if readonly:
        config = get_form_config_from_file('ringo', 'ownership.xml',
                                           'ownership-form-read')
    else:
        config = get_form_config_from_file('ringo', 'ownership.xml',
                                           'ownership-form-update')
    return Form(config, item, db,
                csrf_token=csrf_token,
                eval_url=eval_url,
                url_prefix=url_prefix,
                locale=locale,
                translate=translate)


def get_path_to_form_config(filename, app=None, location=None):
    """Returns the path to the given form configuration. The filename
    should be realtive to the default `location` for the configurations
    in the given `app`.

    :file: Name of the form condifguration.
    :app: You can define the name of the application from where the form
    should be loaded. If no name is provied the code will search along
    the inheritance path of the current application and return the first
    form configuration it can find. Returns None if no configuration can
    be found.
    :location: Default is 'views/forms' you can define an alternative location.
    :returns: Absolute path to the configuration file

    """
    if location is None:
        location = "views/forms"
    if app is None:
        for app in get_app_inheritance_path():
            path = get_path_to(os.path.join(location, filename), app)
            if os.path.exists(path):
                return path
        return path
    else:
        return get_path_to(os.path.join(location, filename), app)


def get_form_config(item, formname):
    """Return the configuration for a given form. The function can
    handle usuall filebased form configurations and blobforms. The
    decision if the item is an instance of the a Blobform is done by
    checking the "fid" foreign key to a form definiton. If present we
    assume that the item is a blobform item.

    :item: Item or class for which the want to get the form
    :formname: name of the form which should be returned
    :returns: Formconfig
    """
    if inspect.isclass(item):
        cachename = "%s.%s" % (item.__name__, formname)
        filename = "%s.xml" % item.__tablename__
    else:
        cachename = "%s.%s" % (item.__class__.__name__, formname)
        filename = "%s.xml" % item.__class__.__tablename__
    name = item.__module__.split(".")[0]
    if not CACHE_FORM_CONFIG.get(cachename):
        if hasattr(item, 'fid'):
            config = get_form_config_from_db(item.fid, formname)
        else:
            config = get_form_config_from_file(name, filename, formname)
        CACHE_FORM_CONFIG.set(cachename, config)
    return CACHE_FORM_CONFIG.get(cachename)


def get_form_config_from_file(name, filename, formname):
    """Return the file based configuration for a given form. The
    configuration tried to be loaded from the current application (and
    its origins) first.  If this fails it tries to load it from the
    extension."""
    loaded_config = None
    try:
        path = get_path_to_form_config(filename)
        if os.path.exists(path):
            loaded_config = load(path)
        elif name.startswith("ringo_"):
            loaded_config = load(get_path_to_form_config(filename,
                                                         name, location="."))
    except:
        pass
    # If we can't load the config file, raise an IOError. Hint: Maybe
    # you missed to set the app.base config variable?
    if loaded_config is None:
        raise IOError("Could not load form configuration for %s" % filename)
    return Config(loaded_config).get_form(formname)


def get_form_config_from_db(fid, formname):
    """Return the blobform configuration for a given form."""
    from ringo.model.form import Form
    factory = Form.get_item_factory()
    form = factory.load(fid)
    return Config(parse(form.definition.encode('utf-8'))).get_form(formname)


def get_formbar_css():
    result = get_css_files()
    for filename, content in result:
        formbar_css_filenames.append(filename)
    return result


def get_formbar_js():
    result = get_js_files()
    for filename, content in result:
        formbar_js_filenames.append(filename)
    return result



from pyramid.scaffolds import PyramidTemplate

class BasicRingoTemplate(PyramidTemplate):
    _template_dir = 'basic'
    summary = 'Ringo basic project'


class RingoExtensionTemplate(PyramidTemplate):
    _template_dir = 'extension'
    summary = 'Ringo extension'

    def pre(self, command, output_dir, vars):
        vars['Package'] = vars['package'].capitalize()
        return PyramidTemplate.pre(self, command, output_dir, vars)



"""Extends the UUID-Field to 36 Characters

Revision ID: cab8c02211c6
Revises: 38eb5f63a8a4
Create Date: 2016-04-22 14:30:01.837297

"""

# revision identifiers, used by Alembic.
revision = 'cab8c02211c6'
down_revision = '38eb5f63a8a4'

from alembic import op
import sqlalchemy as sa


UPGRADE = """"""
DOWNGRADE = """"""


def iter_statements(stmts):
    for st in [x for x in stmts.split('\n') if x]:
        op.execute(st)


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('actions', 'uuid',
               existing_type=sa.CHAR(length=32),
               type_=sa.CHAR(length=36),
               existing_nullable=True)
    op.alter_column('forms', 'uuid',
               existing_type=sa.CHAR(length=32),
               type_=sa.CHAR(length=36),
               existing_nullable=True)
    op.alter_column('modules', 'uuid',
               existing_type=sa.CHAR(length=32),
               type_=sa.CHAR(length=36),
               existing_nullable=True)
    op.alter_column('profiles', 'uuid',
               existing_type=sa.CHAR(length=32),
               type_=sa.CHAR(length=36),
               existing_nullable=True)
    op.alter_column('roles', 'uuid',
               existing_type=sa.CHAR(length=32),
               type_=sa.CHAR(length=36),
               existing_nullable=True)
    op.alter_column('usergroups', 'uuid',
               existing_type=sa.CHAR(length=32),
               type_=sa.CHAR(length=36),
               existing_nullable=True)
    op.alter_column('users', 'uuid',
               existing_type=sa.CHAR(length=32),
               type_=sa.CHAR(length=36),
               existing_nullable=True)
    ### end Alembic commands ###
    iter_statements(UPGRADE)


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.alter_column('users', 'uuid',
               existing_type=sa.CHAR(length=36),
               type_=sa.CHAR(length=32),
               existing_nullable=True)
    op.alter_column('usergroups', 'uuid',
               existing_type=sa.CHAR(length=36),
               type_=sa.CHAR(length=32),
               existing_nullable=True)
    op.alter_column('roles', 'uuid',
               existing_type=sa.CHAR(length=36),
               type_=sa.CHAR(length=32),
               existing_nullable=True)
    op.alter_column('profiles', 'uuid',
               existing_type=sa.CHAR(length=36),
               type_=sa.CHAR(length=32),
               existing_nullable=True)
    op.alter_column('modules', 'uuid',
               existing_type=sa.CHAR(length=36),
               type_=sa.CHAR(length=32),
               existing_nullable=True)
    op.alter_column('forms', 'uuid',
               existing_type=sa.CHAR(length=36),
               type_=sa.CHAR(length=32),
               existing_nullable=True)
    op.alter_column('actions', 'uuid',
               existing_type=sa.CHAR(length=36),
               type_=sa.CHAR(length=32),
               existing_nullable=True)
    ### end Alembic commands ###
    iter_statements(DOWNGRADE)



import sys
from pyramid.view import forbidden_view_config
from pyramid.view import notfound_view_config
from pyramid.response import Response

from ringo.lib.renderer import ErrorDialogRenderer


@forbidden_view_config(path_info='/rest')
def rest_forbidden(request):
    body = '{"success": false, "params": {"error": 403}}'
    response = Response(body=body, content_type='application/json')
    response.status = '403 Forbidden'
    return response


@notfound_view_config(path_info='/rest')
def rest_notfound(request):
    body = '{"success": false, "params": {"error": 404}}'
    response = Response(body=body, content_type='application/json')
    response.status = '404 Not Found'
    return response


@notfound_view_config(renderer='/default/error.mako')
def notfound_get(request):
    _ = request.translate
    title = _("Error 404")
    body = _("Sorry, the requested ressource could not be found!")
    renderer = ErrorDialogRenderer(request, title=title, body=body)
    rvalue = {}
    rvalue['dialog'] = renderer.render()
    request.response.status = 404
    return rvalue


@forbidden_view_config(renderer='/default/error.mako')
def forbidden(request):
    _ = request.translate
    title = _("Error 403")
    body = _("Authorisation error! Sorry, you are not allowed"
             " to access to requested ressource!")
    renderer = ErrorDialogRenderer(request, title=title, body=body)
    rvalue = {}
    rvalue['dialog'] = renderer.render()
    request.response.status = 403
    return rvalue
